#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
 This file is owned exclusively by Smart Arcs Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@keychest.net>, 2019
"""
import socket
import sys
import time
from threading import Thread

from keychest_agent.logger import logger
from keychest_agent.config import KCConfig

__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'


def is_service_running():

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # Connect the socket to the port where the server is listening
    server_address = ('localhost', KCConfig.config.internal_socket_2)

    # noinspection PyBroadException
    try:
        sock.connect(server_address)
        return True
    except:
        return False
    finally:
        sock.close()

def get_registration(force=False):
    """
    Connect to InternalComm server and get a registration ID
    :return:
    """

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # Connect the socket to the port where the server is listening
    server_address = ('localhost', KCConfig.config.internal_socket_2)

    try:
        sock.connect(server_address)

        # Send data
        if force:
            sock.sendall("register force".encode())
        else:
            sock.send("register".encode())

        data = sock.recv(4096)
        data_str = data.decode("utf-8")
        if len(data) < 8:
            return None
        else:
            return data_str
    except Exception as ex:
        sys.stderr.write("Error requesting registration, cause: %s" % str(ex))
        return None
    finally:
        sock.close()

class InternalComm(Thread):
    """
    A simple socket server that will listen to requests while running as service
    """
    def __init__(self, agent_object):

        super().__init__()
        self.agent_object = agent_object

    def run(self):
        logger.debug('KeyChest agent - internal TCP server')
        downstream_port = KCConfig.config.internal_socket_2

        bound = False
        soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        while not bound:
            tries = 20
            while tries > 0 and not bound:
                try:
                    soc.bind(('localhost', downstream_port))
                    logger.debug('Internal socket bound. host:{0}, port:{1}'.
                                 format("localhost", downstream_port))
                    bound = True
                except socket.error:
                    soc.close()
                    tries -= 1
                    time.sleep(5)
                    soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            if not bound:
                logger.error("The port %d is used by another process" % downstream_port)
                # and we will keep trying

        # Start listening on socket
        soc.listen(16)  # default somaxconn

        while True:
            try:
                soc.settimeout(0.5)  # we need to interrupt waiting
                conn, addr = soc.accept()

                # noinspection PyBroadException
                try:
                    data_raw = conn.recv(4096)

                    if len(data_raw) == 0:  # connection was closed
                        break
                    buffer_list = None
                    try:
                        # buffer_list.append(data_raw)
                        if buffer_list is None:
                            buffer_list = data_raw
                        else:
                            buffer_list += data_raw
                    except TypeError:
                        logger.info("Received data can't be converted to text - internal socket")
                        pass

                    # noinspection PyBroadException
                    try:
                        data = buffer_list.decode('utf-8')
                    except Exception:
                        data = ""

                    requests = data.split()
                    register = False
                    force = False
                    for _request in requests:
                        if _request.lower() == "register":
                            register = True
                        elif _request.lower() == "force":
                            force = True

                    if register:
                        if KCConfig.config.agent_email.startswith("dummy") or force:
                            # preserve the logging option
                            local_log = KCConfig.config.do_local_log
                            default_cfg = KCConfig.default_config()
                            KCConfig.write_configuration(default_cfg)
                            KCConfig.config.do_local_log = local_log
                            self.agent_object.do_registration()
                            pass

                        conn.sendall(KCConfig.config.agent_email.encode())
                except BaseException as ex:
                    logger.error("Base exception in internal socket server", cause=str(ex))
                    break
                except IOError:
                    break
                finally:
                    conn.close()

            # here we get on accept timeout
            except socket.timeout:
                pass
            except Exception as ex:
                logger.error("Unexpected exception in internal socket", cause=str(ex))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
 This file is owned exclusively by Smart Arcs Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@keychest.net>, 2019
"""
import socket
from keychest_agent.logger import logger
import dns.resolver

__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'


class TlsUtils(object):
    @staticmethod
    def ip_to_int(ip):
        """
        Converts IPv4 to integer.
        123.123.123.123 -> int

        :param ip:
        :return:
        """
        octets = str(ip).split('.')
        if len(octets) != 4:
            raise ValueError('Invalid IPv4 format, 4 octets required')

        res = 0
        for idx, octet in enumerate([int(x) for x in reversed(octets)]):
            res |= octet << (8 * idx)

        return res

    @staticmethod
    def is_ipv4(ip, port=0):
        # noinspection PyBroadException
        try:
            octets = str(ip).split('.')
            if len(octets) != 4:
                return False

            for idx, octet in enumerate([int(x) for x in reversed(octets)]):
                if octet > 255 or octet < 0:
                    return False

            if isinstance(port, int) and 0 <= port < 65536:
                return True
            else:
                return False
        except Exception:
            return False

    @staticmethod
    def int_to_ip(n):
        """
        Converts 32bit integer to the IP
        :param n:
        :return:
        """
        ip = [str((n >> (8 * i)) & 0xff) for i in range(0, 4)]
        return '.'.join(reversed(ip))

    @staticmethod
    def iter_ips(ip_start=None, ip_stop=None, ip_start_int=None, ip_stop_int=None):
        """
        Iterates over the IP range in the random order - not to trigger IDS while scanning.
        :param ip_start:
        :param ip_stop:
        :param ip_start_int:
        :param ip_stop_int:
        :return:
        """
        if ip_start_int is None:
            ip_start_int = TlsUtils.ip_to_int(ip_start)

        if ip_stop_int is None:
            ip_stop_int = TlsUtils.ip_to_int(ip_stop)

        if ip_stop_int < ip_start_int:
            raise ValueError('Invalid IP order, end > start')

        # Simple case - small range, generate sequence and shuffle
        lst = list(range(ip_start_int, ip_stop_int + 1))
        for cur in lst:
            yield TlsUtils.int_to_ip(cur)
        return

    @staticmethod
    def is_ip(hostname):
        """
        Returns true if the hostname is IPv4 or IPv6
        :param hostname:
        :return:
        """
        ipv4 = TlsUtils.is_valid_ipv4_address(hostname)
        if ipv4:
            return True
        return TlsUtils.is_valid_ipv6_address(hostname)

    @staticmethod
    def is_valid_ipv4_address(hostname):
        """
        Returns true if the hostname is IPv4
        :param hostname:
        :return:
        """

        return TlsUtils.is_ipv4(hostname)
        # r = r'^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])$'
        # if re.match(r, hostname):
        #     return True
        # return False

    @staticmethod
    def is_valid_ipv6_address(address):
        """
        Simple IPV6 address validation using hacky socket.inet_pton
        :param address:
        :return:
        """
        try:
            socket.inet_pton(socket.AF_INET6, address)
        except socket.error:  # not a valid address
            return False
        return True

class DNSUtils(object):

    @staticmethod
    def process_zone(zone, domain=""):
        """
        :param zone:
        :type zone: dns.zone.Zone
        :param  domain
        :type domain:  Union[str|None]
        :return:
        """
        name_map = {}
        if domain is None:
            domain = ""
        if (len(domain) > 0) and (domain[0] == '.'):
            domain  =  domain[1:]

        if zone is not None:
            try:
                # first, let's get 'A' and
                for (name, ttl, rdata) in zone.iterate_rdatas('A'):
                    try:
                        if len(domain)  > 0:
                            if name.to_text() == '@':
                                fqdn = domain
                            else:
                                fqdn = name.to_text() + "." + domain
                        else:
                            fqdn = name.to_text()

                        if fqdn in name_map:
                            name_map[fqdn].append(rdata.address)
                        else:
                            name_map[fqdn] = [rdata.address]
                    except Exception as ex:
                        logger.warning("Error parsing DNS zone data - A",
                                       name="%r" % name.to_text(), domain=domain, rdata="%r" % rdata.address,
                                       cause=str(ex))
            except Exception as ex:
                logger.error("Unexpected exception when parsing 'A' records from zone data",
                             cause=str(ex))

            try:
                # second, let's get 'AAAA' and
                for (name, ttl, rdata) in zone.iterate_rdatas('AAAA'):
                    try:
                        if len(domain)  > 0:
                            if name.to_text() == '@':
                                fqdn = domain
                            else:
                                fqdn = name.to_text() + "." + domain
                        else:
                            fqdn = name.to_text()

                        if fqdn in name_map:
                            name_map[fqdn].append(rdata.address)
                        else:
                            name_map[fqdn] = [rdata.address]
                    except Exception as ex:
                        logger.warning("Error parsing DNS zone data - AAAA",
                                       name="%r" % name.to_text(), domain=domain, rdata="%r" % rdata.address,
                                       cause=str(ex))
            except Exception as ex:
                logger.error("Unexpected exception when parsing 'AAAA' records from zone data",
                             cause=str(ex))

            # the last is CNAME
            try:
                # second, let's get 'AAAA' and
                for (name, ttl, rdata) in zone.iterate_rdatas('CNAME'):
                    try:
                        if len(domain) > 0:
                            if name.to_text() == '@':
                                fqdn = domain
                            else:
                                fqdn = name.to_text() + "." + domain
                        else:
                            fqdn = name.to_text()
                        if rdata.to_text() in name_map:
                            if fqdn not in name_map:  # duplicate CNAMEs??
                                name_map[fqdn] = []
                            # copy IP addresses to CNAME record
                            for each_a_record in name_map[rdata.to_text()]:
                                name_map[fqdn].append(each_a_record)
                        else:
                            logger.warning("CNAME doesn't have a valid A/AAAA record",
                                           name=name.to_text())
                    except Exception as ex:
                        logger.warning("Error parsing DNS zone data - CNAME",
                                       name="%r" % name.to_text(), domain=domain, rdata="%r" % rdata.address,
                                       cause=str(ex))
                pass
            except Exception as ex:
                logger.error("Unexpected exception when parsing 'CNAME' records from zone data",
                             cause=str(ex))
        else:
            logger.info("Zone processing called with undefoned zone")

        return name_map

class ControlUtils(object):
    @staticmethod
    def filter_banned(agent, name_map, connect=0):
        """

        :param agent:
        :param name_map:
        :param connect: positive will attempt to connect to IP addresses on the given port
        :return:
        """
        resolved = []
        tested_ips = {}
        _open = False
        for each_name in name_map:
            # noinspection PyBroadException
            for each_ip in name_map[each_name]:
                if not agent.is_banned(each_ip) and len(each_ip) > 0:
                    if each_ip in tested_ips:
                        if tested_ips[each_ip]:
                            resolved.append({'name': each_name, 'ip': each_ip})
                    else:
                        if connect > 0:
                            _open = False
                            # noinspection PyBroadException
                            try:
                                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                                sock.settimeout(3)
                                result = sock.connect_ex((each_ip, 443))
                                if result == 0:
                                    _open = True
                                sock.close()
                            except:
                                pass
                        else:
                            _open = True
                        tested_ips[each_ip] = _open
                        if _open:
                            resolved.append({'name': each_name, 'ip': each_ip})
                        else:
                            resolved.append({'name': each_name, 'ip': None})

                else:
                    logger.debug("Found a banned IP address", ip=each_ip, name=each_name)
                    pass  # IP address is banned
            pass  # and of the name cycle

        return resolved



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
 This file is owned exclusively by Smart Arcs Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
"""
__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@smartarchitects.co.uk'
__status__ = 'Development'

import requests
import json
import os
import socket
import sys
import threading
from datetime import datetime
from time import time, sleep
import traceback

from keychest_agent.core import Core
from keychest_agent.config import KCConfig
from keychest_agent.logger import logger


class LogDispatcher(threading.Thread):

    def __init__(self, stop_event, sock_file='logbook.sock', ip=None, port=0, to_file=False):
        """
        :param stop_event:
        :param sock_file: the TCP .sock file for listening
        :param ip: or we can listen on an interface and ...
        :param port: ... tcp port
        """
        super().__init__()
        self.stopping = False
        self.messages = []
        self.stop_event = stop_event  # type: threading.Event
        self.stopped = False
        self.to_file = to_file

        if ip is not None:
            self.sock_address = (ip, port)
            self.sock_type = socket.AF_INET
        else:
            self.sock_type = socket.AF_UNIX
            self.sock_address = \
                os.path.join(KCConfig.config.sockdir, sock_file)
            # first we need to create a .sock file for internal communication
            try:
                os.unlink(self.sock_address)
            except OSError:
                if os.path.exists(self.sock_address):
                    raise

        self.code_dir = Core.get_custom_config_file_path('etc/keychest_agent/log_codes.pickle')
        # first, let's empty the queue
        self.myindex = None

    def tcp_server(self):
        """
        This is internal TCP  server for logging - it listens on localhost
        :return:
        """

        while True:
            serv_socket = socket.socket(self.sock_type, socket.SOCK_STREAM)

            tries = 20
            bound = False
            while tries > 0 and not bound:
                try:
                    serv_socket.bind(self.sock_address)
                    sys.stderr.write('Logging socket listening.' + os.linesep)
                    sys.stderr.flush()

                    bound = True
                except socket.error as msg:
                    sys.stderr.write('Cannot bind to logging socket: %s' % str(msg)  + os.linesep)
                    sys.stderr.flush()
                    tries -= 1
                    sleep(5)

            if not bound:
                serv_socket.close()
                continue
            try:
                serv_socket.listen(1)
                while True:
                    # Wait for a connection
                    connection, client_address = serv_socket.accept()
                    end_of_data = False
                    data = ""
                    try:
                        while not end_of_data:
                            data += connection.recv(10000).decode()  # type: str
                            if data:
                                # process data - look for end of line
                                _lines = data.split(os.linesep)
                                # if data ends with os.linesep -> list item will be empty string
                                length = 0
                                for _line in _lines:
                                    length = len(_line)
                                    if length > 0:
                                        self.messages.append(_line)
                                if length > 0:
                                    data = _lines[len(_lines) - 1]  # store the last one
                                else:
                                    data = ''
                            else:
                                # no more data
                                end_of_data = True
                    finally:
                        connection.close()
            except Exception as ex:
                tb = "; ".join(traceback.format_exc(3).splitlines())
                sys.stderr.write("Exception in listening to logging port %s - %s" % (str(ex), tb)  + os.linesep)
                sys.stderr.flush()
            finally:
                serv_socket.close()
                sleep(5)

    def run(self):
        """

        :return:
        """

        # first we need to start the TCP server to accept logs
        tcp_server = threading.Thread(target=self.tcp_server, args=())
        tcp_server.name = "TCPHandler"
        tcp_server.daemon = True  # let it terminate with the main thread
        tcp_server.start()

        # splunk - not now

        # mysocket = None
        # splunk_port = 8089
        # splunk_host = 'localhost'
        # splunk_password = 'password'
        # splunk_user = 'admin'
        # try:
        #     KCConfig.config.logging_host = urllib.parse.quote_plus(socket.gethostname().replace(".", "_"))
        #     loggin_splunk = KCConfig.config.logging_splunk
        #     connection_details = loggin_splunk.split(",")
        #     if len(connection_details) == 1:
        #         splunk_password = connection_details[0]
        #     elif len(connection_details) < 3:
        #         splunk_user = connection_details[0]
        #         splunk_password = connection_details[1]
        #     elif len(connection_details) < 4:
        #         splunk_user = connection_details[0]
        #         splunk_password = connection_details[1]
        #         splunk_host = connection_details[2]
        #     elif len(connection_details) > 3:
        #         splunk_user = connection_details[0]
        #         splunk_password = connection_details[1]
        #         splunk_host = connection_details[2]
        #         try:
        #             splunk_port = int(connection_details[3])
        #         except ValueError:
        #             pass
        #     splunk_ep = sclient.connect(
        #         host=splunk_host,
        #         port=splunk_port,
        #         username=splunk_user,
        #         password=splunk_password
        #     )
        #
        #     existing_indexes = splunk_ep.indexes
        #     create = True
        #     for index in existing_indexes:
        #         if index.name == KCConfig.config.logging_index:
        #             create = False
        #     if create:
        #         splunk_ep.indexes.create(KCConfig.config.logging_index)
        #
        #     self.myindex = splunk_ep.indexes[KCConfig.config.logging_index]
        #     mysocket = self.myindex.attach(sourcetype='keychest', host=KCConfig.config.logging_host)
        #
        # except Exception as ex:
        #     # splunk_ep = None
        #     sys.stdout.write("Can't open splunk connection %s" % str(ex))
        #     sys.stdout.flush()

        agent_email = KCConfig.config.agent_email  # type: str
        if agent_email.startswith("dummy"):
            registered = False
        else:
            registered = True

        logger.load_codes(self.code_dir)
        last_submit = time()
        upload_string = []  # this would be "" for splunk!
        # last_reconnect = time()
        usage_logs_counter = 0
        last_config_test = time()
        log_directory = KCConfig.config.logdir
        log_filename = datetime.now().strftime("agent_%Y_%m_%d.log")
        log_file = open(os.path.join(log_directory, log_filename), 'a')
        while True:
            # noinspection PyBroadException
            try:
                # if not registered:  -- we can de-register and re-register on the fly

                if time() - last_config_test > 10:  # retry no more than every 10 seconds
                    if agent_email != KCConfig.config.agent_email:
                        logger.info("Agent ID has changed", old=agent_email, new=KCConfig.config.agent_email)
                        agent_email = KCConfig.config.agent_email  # type: str

                    if not agent_email.startswith("dummy"):
                        registered = True
                    else:
                        registered = False
                    last_config_test = time()

                if len(self.messages) > 0:
                    record = self.messages.pop(0)  # take the oldest message
                    skip = False
                    # noinspection PyBroadException
                    try:
                        raw = json.loads(record)
                        level = raw['level'].upper()
                        # flag whether it's a trace or usage log message
                        if 'usage' in raw:
                            usage = raw['usage']
                        else:
                            usage = False
                        if usage:
                            level = 'USE'
                            usage_logs_counter = (usage_logs_counter + 1) % KCConfig.config.logging_usage_sampling
                            skip = (usage_logs_counter != 0)

                        if 'params' in raw:
                            params = raw['params']
                        else:
                            params = "{}"

                        if 'code' in raw:
                            code, msg_id = logger.get_code(level, raw['code'], raw['message'])
                        else:
                            code, msg_id = logger.get_code(level, None, raw['message'])

                        _time = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S.%f")
                        _params = json.loads(params)
                        if len(self.messages) > 50:
                            _params['backlog'] = len(self.messages)
                        _message = {'time': raw['time'][:19],
                                    'agent': agent_email,
                                    'event':
                                        {
                                            'level': level,
                                            'code': code,
                                            'message': raw['message'],
                                            'params': _params,
                                            'process': raw['process'],
                                            'id': msg_id,
                                            'host': raw['host'],
                                            'usage': usage,
                                            'git': KCConfig.config.git_version,
                                            'log_time': _time,
                                        }
                                    }
                        message = json.dumps(_message) + os.linesep
                        # if logging_keychest and (time() - last_reconnect > 60):
                        #     if mysocket is None or self.myindex is None and (time() - last_reconnect > 60):
                        #         # don't try to reconnect more often than once a minute
                        #         try:
                        #             last_reconnect = time()
                        #             splunk_ep = sclient.connect(
                        #                 host=splunk_host,
                        #                 port=splunk_port,
                        #                 username=splunk_user,
                        #                 password=splunk_password
                        #             )
                        #             if self.myindex is None:
                        #                 self.myindex = splunk_ep.indexes[KCConfig.config.logging_index]
                        #             if mysocket is None:
                        #                 mysocket = self.myindex.attach(sourcetype='keychest', host=KCConfig.config.logging_host)
                        #         except Exception as ex:
                        #             # splunk_ep = None
                        #             sys.stdout.write("Can't open splunk connection %s" % str(ex))
                        #             sys.stdout.flush()
                        #     pass

                        # if (not skip) and (level in KCConfig.config.logging_remote) and self.myindex and mysocket:
                        if (not skip) and (level in KCConfig.config.logging_remote) and registered:
                            resp = None
                            try:
                                upload_string.append(_message)  # we need JSON here
                                if time() - last_submit > 1:  # every 1 seconds
                                    data = {
                                        'agent': agent_email,
                                        "timestamp": int(time()),
                                        "security": "none",
                                        "signature": None,
                                        "logs": upload_string
                                    }
                                    # add customer-defined items
                                    if isinstance(KCConfig.config.logging_params, dict):
                                        for key, param in KCConfig.config.logging_params.items():
                                            if key not in data:
                                                data[key] = param

                                    headers = {'Accept': 'application/json', 'Content-Type': 'application/json' }
                                    resp = requests.post(KCConfig.config.logging_server, json=data,
                                                         headers=headers, timeout=KCConfig.config.rest_timeout)
                                    last_submit = time()
                                    upload_string = []
                            except Exception as ex:
                                cause = str(ex)
                                code = 0 if resp is None else resp.status_code
                                sys.stderr.write("INFO: Error posting logs to a remote server %s - %s, code: %r" %
                                                 (KCConfig.config.logging_server,  cause, code) + os.linesep)
                                sys.stderr.flush()
                                upload_string = []
                                # self.myindex = None
                                # mysocket = None

                        # local log
                        if KCConfig.config.do_local_log and (level in KCConfig.config.logging_local):
                            if self.to_file:
                                name_now = datetime.now().strftime("agent_%Y_%m_%d.log")
                                # a new day - delete files older than 30 days
                                if name_now != log_filename:
                                    log_filename = name_now
                                    log_file.close()
                                    log_file = open(os.path.join(log_directory, log_filename), 'a')
                                    _, _, files = next(os.walk(log_directory))
                                    to_delete = len(files) - KCConfig.config.log_file_number
                                    while to_delete > 0:
                                        to_delete -= 1
                                        name = files.pop(0)
                                        # noinspection PyBroadException
                                        try:
                                            os.remove(os.path.join(log_directory, name))
                                        except:
                                            pass

                                log_file.write(message)
                                log_file.flush()
                                pass
                            else:
                                sys.stdout.write(message)
                                sys.stdout.flush()

                    except Exception:
                        pass
                else:  # i.e., the queue is empty
                    sleep(2.0)
                    if self.stop_event.is_set():
                        break
            except Exception as ex:
                sys.stderr.write("Exception " + str(ex))
                pass
        # noinspection PyUnreachableCode
        sys.stderr.write("Terminating log dispatcher" + os.linesep)
        self.stopped = True

    def stop(self):
        self.stop_event.set()
        while not self.stopped:
            sleep(0.5)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
 This file is owned exclusively by Smart Arcs Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@keychest.net>, 2019
"""
__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'

try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil
    __path__ = pkgutil.extend_path(__path__, __name__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module: This is a simple port-forward / proxy, written using only the default python
library. If you want to make a suggestion or fix something you can contact-me
at voorloop_at_gmail.com
Distributed over IDC(I Don't Care) license
***

 Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
 This file is owned exclusively by Smart Arcs Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@smartarchitects.co.uk>, May 2018
"""
__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@smartarchitects.co.uk'
__status__ = 'Development'


import socket
from select import select
import sys
import threading
import traceback

from keychest_agent.logger import logger
from keychest_agent.misc import TlsUtils

# Changing the buffer_size and delay, you can improve the speed and bandwidth.
# But when buffer get to high or delay go too down, you can broke things
buffer_size = 4096
delay = 0.0001
forward_to = ('a3.keychest.net', 443)


# A tee for TCP, similar to `socal -v`.
#
#           | server
# client ---|
#           | stdout

class TcpTee:

    def __init__(self, source, destination, job_id, send_up=None, reverse=True):
        """

        :param source:
        :type source: tuple
        :param destination:
        :type destination: tuple
        :param job_id: id of the job
        :type job_id: str
        :param send_up: data to send to the source at the very beginning
        :type send_up: bytearray
        :param reverse: optional data to send to the upstream server
        """
        self.destination = destination  # a tuple, the first item is a list of ip addresses
        self.source = source
        self.send_up = send_up
        self.reverse = reverse
        self.job_id = job_id
        self.clientsock = None
        self.last_channel = None

        # teesock - upstream socket
        if reverse:
            self.teesock = None
        else:
            self.teesock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.teesock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.teesock.bind(source)
            self.teesock.listen(200)

        # Linked client/server sockets in both directions
        self.channel = {}

    def run(self):
        # if reverse -> we connect to upstream and some stuff up there

        if self.reverse:
            ips, port = self.destination
            next_address = None
            initial_done = False
            while len(ips) > 0:
                if next_address:
                    ips.pop(next_address)
                    each_ip = next_address
                else:
                    each_ip = ips.pop(0)
                self.destination = (each_ip, port)
                clientsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                try:
                    clientsock.settimeout(5)
                    clientsock.connect(self.source)
                except Exception as ex:
                    tb = "; ".join(traceback.format_exc(3).splitlines())
                    logger.error('Could not connect to upstream server',
                                 target=self.source, cause=str(ex), traceback=tb, job=self.job_id)

                else:
                    serversock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    try:
                        serversock.settimeout(5)
                        serversock.connect(self.destination)
                    except Exception as ex:
                        tb = "; ".join(traceback.format_exc(3).splitlines())
                        logger.error('Could not connect to server. Closing connection to client as well',
                                     target=self.destination, cause=str(ex), traceback=tb, job=self.job_id)
                        clientsock.close()
                    else:
                        logger.debug("Target connected", ip=self.source[0], job=self.job_id)

                        if not initial_done:
                            if self.send_up:
                                length = len(self.send_up)
                                _len = [0, 0, 0, 0]
                                _len[0] = length % 256
                                _len[1] = (length // 256) % 256
                                _len[2] = (length // (256*256)) % 256
                                _len[3] = (length // (256*256*256)) % 256
                                clientsock.sendall(bytes(_len))
                                clientsock.sendall(self.send_up)
                            else:
                                clientsock.sendall(bytes([0, 0, 0, 0]))
                            logger.debug('Upstream data sent', job=self.job_id)
                            initial_done = True

                        self.channel[clientsock] = serversock
                        self.channel[serversock] = clientsock
                        self.clientsock = clientsock

                        next_address = self.do_proxy()
                        if not next_address and (len(ips) > 0):
                            logger.error("do_proxy should return the next IP address",
                                         job=self.job_id, )
                            break
                        elif next_address in ips:
                            logger.debug("All good, we continue with the next IP address", job=self.job_id)
                            continue
                        elif next_address is None:
                            break
                        else:  # nont None and not in the list
                            logger.error("do_proxy returns a wrong IP address, not present in the list", job=self.job_id)
                            continue

        else:  # not reverse - we wait
            # noinspection PyUnusedLocal
            next_address = self.do_proxy()
            pass

        if self.teesock:
            # noinspection PyBroadException
            try:
                self.teesock.close()
            except:
                pass
        return  # end of the thread

    def do_proxy(self):
        """
        This creates one proxy "session" and terminates when it receives
        7x \x00 + Union[\x00, \x01] + 16x address with padding zeros from the client/upstream
        :return:
        """

        next_address = None
        try:
            while True:
                if self.teesock:
                    select_list = [self.teesock] + list(self.channel)
                else:
                    select_list = list(self.channel.values())
                # readable, writeable, exceptions - parameters are - inputs, outputs, inputs, timeout
                inputready, outputready, exceptready = select(select_list, [], select_list, 500)

                if not (inputready or outputready or exceptready):
                    # timeout
                    self.on_close(self.clientsock)
                    logger.debug("Timeout, no data received on any socket", job=self.job_id)
                    return None

                for s in inputready:
                    if s == self.teesock:
                        self.on_accept()
                        continue
                    data = s.recv(4096)
                    if not data:
                        self.on_close(s)  # shall we check other possible inputs? probably not needed
                        logger.debug("No data received = socket closed, exiting do_proxy",
                                     name=s.getsockname(), job=self.job_id)
                        return None
                    else:
                        _name = s.getsockname()
                        if isinstance(_name, tuple):
                            _name = _name[0]
                        logger.debug("Data received bytes", name=_name, count=len(data))

                    # check if we switched direction and the new one comes from upstream (-> we may need to terminate)
                    if s != self.last_channel:
                        self.last_channel = s
                        if s == self.clientsock and (data[:7] == bytes([0, 0, 0, 0, 0, 0, 0])):
                            # we are terminating this proxy
                            if data[7] == 0:
                                # we are getting IP4 address
                                next_address = 0
                                for i in range(4):
                                    next_address = next_address * 256 + data[8+i]
                                pass
                                logger.debug("Proxy received request to re-connect to a new address",
                                             ip=TlsUtils.int_to_ip(next_address), job=self.job_id)
                            elif data[7] == 1:
                                # we are getting IP6 address
                                next_address = 0
                                for i in range(16):
                                    next_address = next_address * 256 + data[8+i]
                                logger.error("Requested IPv6 address - this is not yet supported",
                                             job=self.job_id, ip=next_address)
                                pass
                            else:
                                # we terminate
                                next_address = None
                                pass
                            self.on_close(s)
                            logger.debug("Received request for re-connect, terminating do_proxy",
                                         job=self.job_id)
                            return next_address

                    self.on_recv(s, data)
                if len(exceptready) > 0:
                    self.on_close(self.clientsock)
                    return None
        finally:
            if self.clientsock:
                self.on_close(self.clientsock)
            return next_address

    def on_accept(self):
        clientsock, clientaddr = self.teesock.accept()
        serversock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            serversock.settimeout(5)
            serversock.connect(self.destination)
        except Exception as ex:
            tb = "; ".join(traceback.format_exc(3).splitlines())
            logger.error('Could not connect to server. Closing connection to client as well',
                         target=self.destination, cause=str(ex), traceback=tb)
            clientsock.close()
        else:
            logger.info("KeyChest server connected", name=clientaddr, job=self.job_id)
            self.channel[clientsock] = serversock
            self.channel[serversock] = clientsock
            self.clientsock = clientsock

    def on_close(self, sock):

        logger.info("Disconnected from target server", target=sock.getpeername(), job=self.job_id)
        othersock = self.channel[sock]

        sock.close()
        othersock.close()

        del self.channel[sock]
        del self.channel[othersock]
        self.clientsock = None
        self.last_channel = None

    def on_recv(self, sock, data):
        self.channel[sock].send(data)


class UdpTee:

    def __init__(self, source, destination, job_id, send_up=None, reverse=True):
        """

        :param source:
        :type source: tuple
        :param destination:
        :type destination: tuple
        :param job_id:
        :type job_id: str
        :param send_up:
        :type send_up: bytearray
        :param reverse:
        :type reverse: bool
        """
        self.destination = destination
        self.source = source
        self.send_up = send_up
        self.reverse = reverse
        self.job_id = job_id

        self.sock_src = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock_dst = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock_src.bind(source)

    def run(self):
        while True:
            data, addr = self.sock_src.recvfrom(65565)
            if not data:
                logger.error('an error occured')
                break
            logger.debug('Received UDP data and forwarded', count=len(data), name=addr, job=self.job_id)
            self.sock_dst.sendto(data, self.destination)
            data, addr2 = self.sock_dst.recvfrom(65565)
            logger.debug('received UDP data', count=len(data), name=addr2, job=self.job_id)
            self.sock_src.sendto(data, addr)

        self.sock_src.close()
        self.sock_dst.close()


class AgentProxy(threading.Thread):

    def __init__(self, protocol, upstream, downstream, job_id, send_up=None, reverse=True):
        """
        This will open a new proxy thread. Any logs produced will be sent via the control channel.

        :param protocol: it must be "tcp" or "udp"
        :param upstream: a tuple (address, port) of the KeyChest server
        :type upstream: tuple
        :param downstream: a tuple (address, port) of the audit target
        :type downstream: tuple
        :param job_id: name of the job being processed
        :type job_id: str
        :param send_up: bytearray to send to the upstream (KeyChest server)
        :type send_up: bytearray
        :param reverse:  reverse proxy or not
        """
        super().__init__()
        self.reverse = reverse
        if protocol not in ['tcp', 'udp']:
            raise Exception()

        self.protocol = protocol
        self.destination = downstream
        self.upstream = upstream
        self.send_up = send_up
        self.job_id = job_id
        self.tee = None

    def run(self):
        if self.protocol not in ['tcp', 'udp']:
            logger.error("Unknown proxy protocol", name=self.protocol, job=self.job_id)
            return

        try:
            if self.protocol == 'tcp':
                port = self.destination[1]
                ips = self.destination[0]
                if isinstance(ips, str):
                    ips = [ips]
                self.tee = TcpTee(self.upstream, (ips, port), self.job_id, send_up=self.send_up,
                                  reverse=self.reverse)
            else:
                self.tee = UdpTee(self.upstream, self.destination, self.job_id, send_up=self.send_up,
                                  reverse=self.reverse)
            self.tee.run()
            logger.debug("Proxy thread finished", job=self.job_id)
        except Exception as ex:
            tb = "; ".join(traceback.format_exc(3).splitlines())
            logger.error("Proxy thread finished with exception", job=self.job_id, traceback=tb, cause=str(ex))
            pass
        finally:
            sys.exit(1)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-listen_port", default=11111, help="The port this process will listen on.", type=int)
    # parser.add_argument("-server_host", default="a3.keychest.net", help="The remote host to connect to.")
    # parser.add_argument("-server_port", default=443, help="The remote port to connect to.", type=int)
    parser.add_argument("-server_host", default="131.111.12.20", help="The remote host to connect to.")
    parser.add_argument("-server_port", default=53, help="The remote port to connect to.", type=int)
    args = parser.parse_args()

    #  logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)

    # tee = TcpTee(int(args.listen_port), (args.server_host, int(args.server_port)))
    tee = UdpTee(("127.0.0.1", args.listen_port), (args.server_host, int(args.server_port)), "tj1")
    try:
        tee.run()
    except KeyboardInterrupt:
        logger.info("Ctrl C - Good Bye")
        sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
 This file is owned exclusively by Smart Arcs Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@smartarchitects.co.uk>, May 2018
"""
import os
import sys
import json
import collections
import errno
import time
from datetime import datetime
from keychest_agent.core import Core

from keychest_agent.logger import logger

__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@smartarchitects.co.uk'
__status__ = 'Development'


class KCConfig(object):
    """Configuration object, handles file read/write"""

    # let's create a static version - default file
    config = None

    CONFIG_FILE = 'agent_config.json'

    # noinspection PyUnusedLocal
    def __init__(self, json_db=None, *args, **kwargs):
        self.json = json_db

    def ensure_config(self):
        if self.json is None:
            self.json = collections.OrderedDict()
        if 'config' not in self.json:
            self.json['config'] = collections.OrderedDict()

    def has_nonempty_config(self):
        return self.json is not None and 'config' in self.json and len(self.json['config']) > 0

    def get_config(self, key, default=None):
        if not self.has_nonempty_config():
            logger.warning("Request for non-existent configuration option, returning default value",
                           name=key, value=default)
            return default
        return self.json['config'][key] if key in self.json['config'] else default

    @classmethod
    def get_static_config(cls, key, default=None):
        if KCConfig.config is None:
            return default
        return KCConfig.config.json['config'][key] if key in KCConfig.config.json['config'] else default

    # @staticmethod
    # def init_log(idx=None, level=logging.DEBUG):
    #     """
    #     Initializes logging
    #     :return:
    #     """
    #     log_folder = KCConfig.get_custom_config_file_path('var/log/keychest')
    #     util.make_or_verify_dir(log_folder)
    #
    #     if idx is not None:
    #         file_name = 'server_%d.log' % idx
    #     else:
    #         file_name = 'server.log'
    #     _logger = QueueHandler.set_logger(os.path.join(log_folder, file_name))
    #     _logger.setLevel(level)
    #     return _logger

    def set_config(self, key, val):
        self.ensure_config()
        self.json['config'][key] = val

    def to_string(self):
        return json.dumps(self.json, indent=2) if self.has_nonempty_config() else ""

    @property
    def log_file_number(self):
        return self.get_config('log_file_number', 30)

    @property
    def logging_usage_sampling(self):
        return self.get_config("logging_usage_sampling", 1)

    @property
    def do_register(self):
        return self.get_config('do_register', False)

    @do_register.setter
    def do_register(self, value):
        self.set_config('do_register', value)

    @property
    def logging_server(self):  # this points to Laravel
        return self.get_config('logging_server', 'https://keychest.net/api/v1.0/agent/log')

    @logging_server.setter
    def logging_server(self, value):
        self.set_config('logging_server', value)

    @property
    def agent_name(self):
        return self.get_config('agent_name', "KEYCHEST_" + datetime.utcnow().strftime("%Y%m%d%H%M"))

    @agent_name.setter
    def agent_name(self, value):
        self.set_config('agent_name', value)

    @property
    def logging_params(self):
        return self.get_config('logging_params', {})

    @property
    def control_server(self):  # this points to Engine's RESTful server
        return self.get_config('control_server', 'http://keychest.net:8443/api/v1.0/agent/control')

    @control_server.setter
    def control_server(self, value):
        self.set_config('control_server', value)

    @property
    def agent_email(self):
        return self.get_config('agent_email', "dummy@keychest.net")

    @agent_email.setter
    def agent_email(self, value):
        self.set_config('agent_email', value)

    @property
    def logging_remote(self):
        return self.get_config('logging_remote', ["ERROR", "INFO", "CRITICAL", "WARNING", "DEBUG", "USE"])

    @logging_remote.setter
    def logging_remote(self, value):
        self.set_config('logging_remote', value)

    @property
    def do_local_log(self):
        return self.get_config('do_local_log', False)

    @do_local_log.setter
    def do_local_log(self, value):
        self.set_config('do_local_log', value)

    @property
    def logging_local(self):
        return self.get_config('logging_local', ["ERROR", "INFO", "CRITICAL", "WARNING", "DEBUG", "USE", "TRACE"])

    @logging_local.setter
    def logging_local(self, value):
        self.set_config('logging_local', value)

    # The number of processes for audit engine
    @property
    def agent_processes(self):
        return self.get_config('agent_processes', 1)

    @agent_processes.setter
    def agent_processes(self, value):
        if isinstance(value, int) and (value > 0):
            self.set_config('agent_processes', value)

    @property
    def simple_tls(self):
        return self.get_config('simple_tls', True)

    @simple_tls.setter
    def simple_tls(self, value):
        if isinstance(value, bool):
            self.set_config('simple_tls', value)

    @property
    def rest_timeout(self):
        return self.get_config('rest_timeout', 5)

    @rest_timeout.setter
    def rest_timeout(self, value):
        if isinstance(value, int) and (value > 0) and (value < 60):
            self.set_config('rest_timeout', value)

    @property
    def git_version(self):
        return self.get_config('git_version', "unknown")

    @git_version.setter
    def git_version(self, value):
        if isinstance(value, bytes):
            value = value.decode().strip()
        self.set_config('git_version', value)

    # The max time for each process to run
    @property
    def process_running_time(self):
        return self.get_config('process_running_time', 0)

    # API key for master
    @property
    def agent_apikey(self):
        return self.get_config('agent_apikey', None)

    @agent_apikey.setter
    def agent_apikey(self, val):
        self.set_config('agent_apikey', val)

    @property
    def max_process_memory(self):
        return self.get_config('max_process_memory', 200)

    @max_process_memory.setter
    def max_process_memory(self, value):
        if isinstance(value, int):
            self.set_config('max_process_memory', value)

    @property
    def local_config(self):
        return self.get_config('local_config', False)

    @local_config.setter
    def local_config(self, value):
        if isinstance(value, bool):
            self.set_config('local_config', value)

    @property
    def reverse_proxy(self):
        return self.get_config('reverse_proxy', True)

    @reverse_proxy.setter
    def reverse_proxy(self, value):
        if isinstance(value, bool):
            self.set_config('reverse_proxy', value)

    @property
    def dns_server(self):
        temp = self.get_config('dns_server', None)
        if (temp is None) or (len(temp) < 1):
            return None
        else:
            return temp

    @dns_server.setter
    def dns_server(self, value):
        if isinstance(value, str) or (value is None):
            self.set_config('dns_server', value)

    @property
    def upstream_permitted_host(self):
        temp = self.get_config('upstream_permitted_host', None)
        if (temp is None) or (len(temp) < 1):
            return None
        else:
            return temp

    @upstream_permitted_host.setter
    def upstream_permitted_host(self, value):
        if isinstance(value, str) or (value is None):
            self.set_config('upstream_permitted_host', value)

    @property
    def agent_proxy_ports(self):
        return self.get_config('agent_proxy_ports', "10001")

    @agent_proxy_ports.setter
    def agent_proxy_ports(self, value):
        if isinstance(value, str):
            parts = value.split('-')
            if len(parts) > 3:
                err = False
                for x in parts:
                    try:
                        v = int(x)
                        if (v < 1023) or (v > 65535):
                            err = True
                            break
                    except ValueError:
                        err = True
                        break
                if not err:
                    self.set_config('agent_proxy_ports', value)
        elif value is None:
            self.set_config('agent_proxy_ports', value)

    @property
    def audit_server_port(self):
        return self.get_config('audit_server_port',  10000)

    @audit_server_port.setter
    def audit_server_port(self, value):
        self.set_config('audit_server_port', value)

    @property
    def upstream_host(self):
        return self.get_config('upstream_host', 'keychest.net')

    @upstream_host.setter
    def upstream_host(self, value):
        self.set_config('upstream_host', value)

    @property
    def agent_command_list(self):
        return self.get_config('agent_command_list', None)

    @agent_command_list.setter
    def agent_command_list(self, value):
        self.set_config('agent_command_list', value)

    @property
    def job_schedule(self):
        return self.get_config('job_schedule', 120)  # seconds

    @property
    def control_schedule(self):
        return self.get_config('control_schedule', 20)

    @control_schedule.setter
    def control_schedule(self, value):
        if isinstance(value, int) and (value >= 20) and (value <= 36000):
            self.set_config('control_schedule', int(value))

    @property
    def discovery_schedule(self):
        return self.get_config('discovery_schedule', 7200)

    @discovery_schedule.setter
    def discovery_schedule(self, value):
        if isinstance(value, int) and (value >= 60) and (value <= 604800):
            self.set_config('discovery_schedule', int(value))

    @property
    def permitted_space(self):
        return self.get_config('permitted_space', "0.0.0.0/0:0")

    @permitted_space.setter
    def permitted_space(self, value):
        self.set_config('permitted_space', value)

    @property
    def banned_space(self):
        return self.get_config('banned_space', '')

    @banned_space.setter
    def banned_space(self, value):
        self.set_config('banned_space', value)

    @property
    def throttling(self):
        return self.get_config('throttling')

    @throttling.setter
    def throttling(self, value):
        self.set_config('throttling', value)

    @property
    def piddir(self):
        return KCConfig.get_custom_config_file_path('var/run/keychest')

    @property
    def logdir(self):
        return KCConfig.get_custom_config_file_path('var/log/keychest')

    @property
    def sockdir(self):
        return KCConfig.get_custom_config_file_path('var/sock')

    @property
    def rundir(self):
        return KCConfig.get_custom_config_file_path('var/keychest/target')

    @property
    def discovery_threads(self):
        return self.get_config('discovery_threads', 1)

    @discovery_threads.setter
    def discovery_threads(self, value):
        if isinstance(value, int) and 0 < value < 100:
            self.set_config("discovery_threads", value)

    @property
    def default_port(self):
        return self.get_config("default_port", 443)

    @property
    def internal_socket(self):
        return self.get_config("internal_socket", 9999)

    @property
    def internal_socket_2(self):
        return self.internal_socket - 1

    @default_port.setter
    def default_port(self, value):
        if isinstance(value, int) and 0 < value < 65536:
            self.set_config('default_port', value)

    @property
    def connect_timeout(self):
        return self.get_config("connect_timeout", 4)

    @connect_timeout.setter
    def connect_timeout(self, value):
        if isinstance(value, int) and value > 0:
            self.set_config("connect_timeout", value)

    @property
    def ad_password(self):
        return self.get_config('ad_password', "")

    @property
    def ad_user(self):
        return self.get_config('ad_user', "")

    local_items = {
        'ad_user': "<set locally>",
        'ad_password': "<set locally>",
        'agent_apikey': "%5.5s  ...",
        'log_file_number': "<internal>",
        'do_register': "<internal>",
        'agent_processes': '<internal>',
        'process_running_time': '<internal>',
        'logdir': '<internal>',
        'piddir': '<internal>',
        'sockdir': '<internal>',
        'rundir': '<internal>'
    }

    @classmethod
    def mask_local_item(cls,  key, value):

        if key in KCConfig.local_items:
            # noinspection PyBroadException
            try:
                return KCConfig.local_items[key] % value
            except:
                return KCConfig.local_items[key]
        else:
            return value

    @classmethod
    def is_local_item(cls, key):
        return key in KCConfig.local_items

    @classmethod
    def create_config_dict(cls, config, hide_local=True):

        result = None
        try:
            result = json.loads(json.dumps(config))
            if hide_local:
                for item, value in result.items():
                    # noinspection PyBroadException
                    try:
                        result[item]  =  KCConfig.mask_local_item(item, value)
                    except Exception:
                        pass
        except Exception as ex:
            logger.error("Exception when loading the agent configuration", cause=str(ex))

        return result

    @classmethod
    def default_config(cls):
        def_cfg = collections.OrderedDict()

        def_cfg['agent_name'] = "KEYCHEST" + datetime.utcnow().strftime("%Y%m%d%H%M")
        def_cfg['agent_apikey'] = ""
        def_cfg['agent_email'] = "dummy@keychest.net"
        def_cfg['logging_server'] = 'https://keychest.net/api/v1.0/agent/log'
        def_cfg['logging_params'] = {}
        def_cfg['control_server'] = 'http://keychest.net:8443/api/v1.0/agent/control'
        def_cfg['audit_server_port'] = 10000
        def_cfg['upstream_host'] = 'keychest.net'
        def_cfg['local_config'] = False  # when True, KeyChest pulls config, otherwise it pushes it to agents
        def_cfg['reverse_proxy'] = True  # when True, audits are done via a reverse proxy
        def_cfg['dns_server'] = ""  # when an IP address / domain name is provided it has to be used
        #  it can contain IP range / domain mask & port, eg 10.0.0.1-10.0.0.3:443
        def_cfg['upstream_permitted_host'] = ""  # if set, only requests from those IP addresses are allowed
        def_cfg['agent_processes'] = 1  # limits the number of concurrent proxies
        def_cfg['agent_proxy_ports'] = 10001  # the starting port that the agent will use for proxies to listen on
        def_cfg['agent_cmd_list'] = ""  # if set, it limits the commands it's allowed to proxy
        def_cfg['max_process_memory'] = 200
        def_cfg['logging_remote'] = ["ERROR", "INFO", "CRITICAL", "WARNING", "DEBUG", "USE"]
        def_cfg['logging_local'] = ["ERROR", "INFO", "CRITICAL", "WARNING", "DEBUG", "TRACE", "USE"]
        def_cfg['simple_tls'] = True
        def_cfg['rest_timeout'] = 5
        def_cfg['control_schedule'] = 120
        def_cfg['discovery_schedule'] = 7200
        def_cfg['permitted_space'] = "0.0.0.0/0:0"
        def_cfg['banned_space'] = ""
        def_cfg['throttling'] = 0
        def_cfg['discovery_threads'] = 1
        def_cfg['default_port'] = 443
        def_cfg['ad_password'] = ""
        def_cfg['ad_user'] = ""
        def_cfg['discovery_threads'] = 10
        def_cfg['connect_timeout'] = 3
    # def_cfg['agent_running_time'] = 0  # it will restart when the time limit is reached

        root = collections.OrderedDict()
        root['config'] = def_cfg
        return cls(json_db=root)

    @classmethod
    def get_config_file_path(cls):
        """Returns basic configuration file"""
        return os.path.join(Core.CONFIG_DIR, KCConfig.CONFIG_FILE)

    @classmethod
    def config_file_exists(cls):
        conf_name = KCConfig.get_config_file_path()
        return os.path.exists(conf_name) and os.path.isfile(conf_name)

    @classmethod
    def is_configuration_nonempty(cls):
        return KCConfig.config is not None and KCConfig.config.has_nonempty_config()

    @classmethod
    def read_configuration(cls, force=False):
        if KCConfig.config and (force is not True):
            return KCConfig.config
        else:
            if not KCConfig.config_file_exists():
                return None

            conf_name = KCConfig.get_config_file_path()
            KCConfig.config = KCConfig.from_file(conf_name)
            return KCConfig.config

    @classmethod
    def write_configuration(cls, cfg):

        try:
            os.makedirs(Core.CONFIG_DIR, mode=0o755)
        except OSError as exception:
            if exception.errno == errno.EEXIST:
                pass
            else:
                raise

        conf_name = KCConfig.get_config_file_path()
        # noinspection PyBroadException
        try:
            os.rename(conf_name, conf_name + "_%d" % int(time.time()))
        except Exception:
            # this is likely the first installation or the configuration file has been deleted
            pass

        with os.fdopen(os.open(conf_name, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o600), 'w') as config_file:
            config_file.write('// \n')
            config_file.write('// Agent config file generated: %s\n' % datetime.utcnow().strftime("%Y-%m-%d %H:%M"))
            config_file.write('// \n')
            config_file.write(cfg.to_string() + "\n\n")
        KCConfig.config = cfg
        return conf_name

    @classmethod
    def from_json(cls, json_string):
        return cls(json_db=json.loads(json_string, object_pairs_hook=collections.OrderedDict))

    @classmethod
    def from_file(cls, file_name):
        with open(file_name, 'r') as f:
            read_lines = [x.strip() for x in f.read().split('\n')]
            lines = []
            for line in read_lines:
                if line.startswith('//'):
                    continue
                lines.append(line)

            my_config = None
            try:
                my_config = KCConfig.from_json('\n'.join(lines))
            except Exception as ex:
                logger.error("Error parsing configuration file", query=file_name, cause=str(ex))
                sys.stderr.write("Error parsing configuration file: " + str(ex))
                sys.stderr.flush()
                exit(-1)

            return my_config
        pass

    @classmethod
    def get_custom_config_file_path(cls, filename):
        """Returns configuration file for my own file"""
        return os.path.join(Core.CONFIG_DIR, filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
# This file is owned exclusively by Smart Arcs Ltd.
# Unauthorized copying of this file, via any medium is strictly prohibited
# Proprietary and confidential
import sys

__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@smartarchitects.co.uk'
__status__ = 'Development'

import os
import os.path


# noinspection PyBroadException
class Core(object):
    """
    Pidlock + configuration
    """

    CONFIG_DIR = '.'
    try:
        CONFIG_DIR = os.environ['KC_CONFIG_DIR']
    except Exception:
        # hopefully just the key is not set
        CONFIG_DIR = os.path.expanduser("~/.keychest")
        try:
            _path = os.path.join(CONFIG_DIR)
            if not os.path.exists(_path):
                os.makedirs(_path, mode=0o755)

        except Exception as ex:
            CONFIG_DIR = os.path.expanduser("~")

    @classmethod
    def clear_pid(cls):
        pid_file = os.path.join(os.path.join(Core.CONFIG_DIR, 'var/run'), "keychest_agent.pid")
        # my_pid = "%s" % os.getpid()
        if os.path.exists(pid_file):
            os.remove(pid_file)

    def __init__(self):
        """Init the core functions"""

        pid_file = os.path.join(os.path.join(Core.CONFIG_DIR, 'var/run'), "keychest_agent.pid")
        my_pid = "%s" % os.getpid()
        if os.path.exists(pid_file):
            with open(pid_file, 'r') as f:
                file_pid = f.read()
            if file_pid != my_pid:
                sys.stderr.write("There is a running agent with process id %s" % open(pid_file, 'r').read())
                exit(-1)
        else:
            if not os.path.exists(os.path.join(Core.CONFIG_DIR, 'var/run')):
                os.makedirs(os.path.join(Core.CONFIG_DIR, 'var/run'), mode=0o755)
            f = open(pid_file, 'w')
            f.write("%r" % os.getpid())
            f.close()
        try:
            _path = os.path.join(Core.CONFIG_DIR, 'var/run')
            if not os.path.exists(_path):
                os.makedirs(_path, mode=0o755)
        except:
            sys.stderr.write("Can't create a config folder '<base_path>/var/run'\n")

        self.pidlock_created = False

    @classmethod
    def get_custom_config_file_path(cls, filename):
        """Returns configuration file for my own file"""
        return os.path.join(Core.CONFIG_DIR, filename)

    # def pidlock_create(self):
    #     """
    #     Creates a new pidlock if it was not yet created
    #     :return:
    #     """
    #     if not self.pidlock_created:
    #         self.pidlock.create()
    #         self.pidlock_created = True

    # def pidlock_check(self):
    #     """
    #     Checks if the current process owns the pidlock
    #     :return: True if the current process owns the pidlock
    #     """
    #     return self.pidlock.check()

    # def pidlock_get_pid(self):
    #     """
    #     Returns pid of the process holding pidlock, None if there is none.
    #     :return:
    #     """
    #     filename = self.pidlock.filename
    #     if filename and os.path.isfile(filename):
    #         try:
    #             with open(filename, "r") as fh:
    #                 fh.seek(0)
    #                 _pid = int(fh.read().strip())
    #                 return _pid
    #         except:
    #             pass
    #
    #     return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import os
import platform
import signal
import socket
import subprocess
import sys
import threading
import time

import logbook as logbook
import pkg_resources

from keychest_agent.internal_comm import InternalComm, get_registration, is_service_running
from keychest_agent.core import Core
from keychest_agent.agent import Agent
from keychest_agent.config import KCConfig
from keychest_agent.log_dispatcher import LogDispatcher
from keychest_agent.logger import logger
from keychest_agent.version import Version

app = None


def parse_args():
    """
    Argument parsing & startup
    :return:
    """
    # Parse our argument list
    parser = argparse.ArgumentParser(description='KeyChest Agent, version %s' % Version.VERSION)

    # noinspection PyTypeChecker
    # parser.add_argument('--logging', dest='logging_level', default='notice', action='store_const', const=True,
    #                     help='possible logging levels are "debug", "info", "notice", "warning", "error", "critical"')

    parser.add_argument('--init', dest='init_config_only', default=False, action='store_const', const=True,
                        help="Creates a default configuration file if none found and stops")
    parser.add_argument('--register', dest='register_only', default=False, action="store_const", const=True,
                        help="Only register with KeyChest services and exit")
    parser.add_argument('--force', dest='force_register', default=False, action="store_const", const=True,
                        help="When used with --register or --getid, it will delete existing configuration file, no effect otherwise")
    parser.add_argument('--getid', dest='get_register', default=False, action="store_const", const=True,
                        help="Requests the agent ID from a running service")
    parser.add_argument('--staging', dest='staging_env', default=False, action='store_const', const=True,
                        help="If set, the configuration file will be pointing to the KeyChest staging environment")
    parser.add_argument('--testing', dest='testing_env', default=False, action='store_const', const=True,
                        help="If set, the configuration file will be pointing to localhost for testing")
    parser.add_argument('--single', dest='not_force_run', default=False, action='store_const', const=True,
                        help="Default is to ignore existing PID file - this option will terminate if a PID file exists")
    parser.add_argument('--filelog', dest='do_file_logging', default=False, action='store_const', const=True,
                        help="Enables local logging to $HOME/.keychest/var/log")
    parser.add_argument('--locallog', dest='do_local_log', default=False, action='store_const', const=True,
                        help="Enables local logging to stdout")
    parser.add_argument('--run', dest='do_run_win', default=False, action='store_const', const=True,
                        help="Helper argument for Windows Service, no actual effect")

    return parser.parse_args()


# noinspection PyUnusedLocal
def signal_handler(sig, frame):
    global app

    Core.clear_pid()
    if app is not None:
        app.log_dispatcher_thread.stop()
    raise SystemExit('Terminating KeyChest agent')

def main():
    """
    Main keychest_agent starter
    :return:
    """
    global app

    # let's setup a config value for the current git version
    resource_package = __name__
    data_filename = pkg_resources.resource_filename(resource_package, "version.py")
    # noinspection PyBroadException
    try:
        git_version_b = subprocess.check_output(['git', 'describe', '--always'], stderr=subprocess.DEVNULL)  # returns bytes
        git_version = git_version_b.decode().strip()

        with open(data_filename, 'w') as git_file:
            git_file.write("\nclass Version(object):\n    VERSION = '" + git_version + "'\n")

    except Exception:  # if we can't run git, read version from a file (created during installation of the package
        pass

    args = parse_args()

    if (threading.current_thread() is threading.main_thread()) and (not args.get_register):
        signal.signal(signal.SIGINT, signal_handler)

    KCConfig.read_configuration()
    if KCConfig.config is None:
        KCConfig.config = KCConfig.default_config()

    if args.get_register:
        name = get_registration(args.force_register)
        if name is None:
            # the service is not running
            sys.stderr.write("\n\n\nService is not running:\n check that KeyChest Agent is up and running "
                             "If it is, contact our support at:\n\n  support@keychest.net\n\n")

        else:
            print_id(name)
        sys.exit(0)
    else:
        # check that no service is running
        if is_service_running():
            sys.stdout.write("Keychest Agent is already running.\n\n")
            sys.exit(0)

    if not args.not_force_run:
        Core.clear_pid()
    Core()  # init and create pid file

    if args.force_register and args.register_only:
        default_cfg = KCConfig.default_config()
        KCConfig.write_configuration(default_cfg)

    app = Agent(args)

    if args.register_only:
        KCConfig.config.do_register = True
    else:
        KCConfig.config.do_register = False
    app.old_stderr = sys.stderr
    KCConfig.config.do_local_log = args.do_local_log or args.do_file_logging
    if not KCConfig.config.do_local_log:
        f = open(os.devnull, 'w')
        sys.stderr = f

    if app.args.staging_env:
        default_cfg = KCConfig.default_config()
        KCConfig.config.logging_server \
            = default_cfg.logging_server.replace("keychest.net", "a3.keychest.net")
        if "a3.keychest.net" not in KCConfig.config.control_server:
            KCConfig.config.control_server \
                = default_cfg.control_server.replace("keychest.net", "a3.keychest.net")
        KCConfig.write_configuration(KCConfig.config)
    elif app.args.testing_env:
        default_cfg = KCConfig.default_config()
        KCConfig.config.logging_server \
            = default_cfg.logging_server.replace("https://keychest.net", "http://127.0.0.1")
        KCConfig.config.control_server \
            = default_cfg.control_server.replace("http://keychest.net", "http://127.0.0.1")
        KCConfig.write_configuration(KCConfig.config)

    # Init
    if app.args.init_config_only:
        Core.clear_pid()
        return

    # noinspection PyBroadException
    try:
        # subscriber = MultiProcessingSubscriber()
        # app.logging_queue = subscriber.queue
        logging_socket = False
        if hasattr(app.args, 'do_file_logging') and app.args.do_file_logging:
            file_output = True
        else:
            file_output = False

        if platform.system() in ['Windows', 'nt']:
            logging_socket = True
            local_ip = socket.gethostbyname("localhost")
            app.log_dispatcher_thread = LogDispatcher(app.logger_stop_event, ip=local_ip,
                                                      port=KCConfig.config.internal_socket, to_file=file_output)
        else:
            app.log_dispatcher_thread = LogDispatcher(app.logger_stop_event, to_file=file_output)

        app.log_dispatcher_thread.name = "Log dispatcher"
        app.log_dispatcher_thread.daemon = True
        app.log_dispatcher_thread.start()
        Agent.setup_logging(logbook.DEBUG, logging_socket)
        time.sleep(1)  # wait a bit for the logger to start listening on TCP
    except:
        Core.clear_pid()
        sys.exit(-1)

    logger.info("Logger configured for multiprocessing with stderr and filehandler")
    # app.logger = KCConfig.init_log()

    # filename = os.path.join(log_folder, 'server.log')
    # if filename:
    #     handlers.append(logbook.FileHandler(filename, mode='a', level=logbook.DEBUG, bubble=True))
    # target_handlers = logbook.NestedSetup(handlers)
    # sub = MultiProcessingSubscriber(app.logging_queue)
    # sub.dispatch_in_background(target_handlers)

    logger.info("KeyChest Agent starting", name=KCConfig.config.agent_name, query=str(app.args))
    if KCConfig.config.local_config:
        logger.info("The configuration is managed locally")
    else:
        logger.info("The configuration is managed from the KeyChest service")

    # create an internal comm server
    com_thread = InternalComm(app)
    com_thread.name = "InternalCom"
    com_thread.start()

    # the main processing loop
    terminate = False
    while not terminate:
        try:
            terminate = app.work_loop()
        except Exception as ex:
            logger.error("Agent main work_loop terminated", cause=str(ex))
            pass

    app.log_dispatcher_thread.stop()
    if KCConfig.config.do_register:
        print_id()

    Core.clear_pid()


def print_id(name=None):
    if name is None:
        sys.stderr = app.old_stderr
        name = KCConfig.config.agent_email
    if name == "dummy@keychest.net":
        sys.stderr.write("\n\n\nRegistration failed:\n check the log messages above and if required, "
                         "send the logs to our support:\n\n  support@keychest.net\n\n")
    else:
        sys.stderr.write("\n\n\nThis agent ID is:\n %s\n\n\n" % name)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
 This file is owned exclusively by Smart Arcs Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@keychest.net>, 2019
"""
import base64
import hashlib
import hmac
import json
import time

from keychest_agent.logger import logger
from keychest_agent.pyaes.aespy import AESModeOfOperationCBC
from keychest_agent.pyaes.blockfeeder import Encrypter, Decrypter

__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'


class Security(object):
    """
    A simple static class to secure and "unsecure" JSON data
    """

    @classmethod
    def secure(cls, data, key):
        """
        It takes a JSON structure, searches for "payload", encrypts it and adds a new
        item "signature" and "iv", which is HMAC of the encrypted data.
        1. check the data
        2. encryption
        3. encrypted string is base64-d
        4. HMAC
        5. HJMAC is based64
        :param data:
        :param key
        :return:
        """

        # check the data
        try:
            _plaintextdata = json.dumps(data['payload']).encode()
            _ = json.loads(_plaintextdata)
        except Exception as ex:
            logger.error("Error checking data for encryption", cause=str(ex))
            return None
        else:
            try:
                # create diversified keys
                raw_key_enc = "ENC" + key
                raw_key_mac = "MAC" + key
                _hash = hashlib.sha256()
                _hash.update(raw_key_enc.encode())
                key_enc = _hash.digest()[:16]

                _hash = hashlib.sha256()
                _hash.update(raw_key_mac.encode())
                key_mac = _hash.digest()[:16]

                # prepare IV - from timestamp
                _hash = hashlib.sha256()
                raw_iv = "%f" % time.time()
                _hash.update(raw_iv.encode())
                iv = _hash.digest()[:16]

                # create an encryptor and encrypt the data
                aes = Encrypter(AESModeOfOperationCBC(key_enc, iv=iv))
                encrypted = aes.feed(_plaintextdata)
                encrypted += aes.feed()
                data_out = base64.b64encode(encrypted)

                # HMAC of the encrypted data
                _hmac = hmac.new(key_mac, data_out, hashlib.sha256)
                _hmaced = _hmac.digest()
                signature_out = base64.b64encode(_hmaced)

                # and prepare the output
                iv_out = base64.b64encode(iv).decode()
                data['payload'] = data_out.decode()
                data['signature'] = {'iv': iv_out, 'hmac': signature_out.decode()}

                return data
            except Exception as ex:
                logger.error("Exception when encrypting data", cause=str(ex))
                return None

    @classmethod
    def extract(cls, data, key):
        """
        It takes a JSON structure and searches for "data" and "signature" if either is not
        found, nothing happens and it returns data intact. If both exist
        1. de-base64 signature
        2. verifies the signature = HMAC
        3. base64 decode of data
        4. data decryption
        :param data:
        :param key:
        :return:
        """

        # check the data
        try:
            _encrypted_data = data['payload']
        except Exception as ex:
            logger.info("Error checking data for decryption", cause=str(ex))
            return None
        else:

            if ('signature' not in data) or ('iv' not in data['signature']) or ('hmac' not in data['signature']):
                logger.info("Data not encrypted or the signature is missing")
                return data

            try:
                # create diversified keys
                raw_key_enc = "ENC" + key
                raw_key_mac = "MAC" + key
                _hash = hashlib.sha256()
                _hash.update(raw_key_enc.encode())
                key_enc = _hash.digest()[:16]

                _hash = hashlib.sha256()
                _hash.update(raw_key_mac.encode())
                key_mac = _hash.digest()[:16]

                # HMAC of the encrypted data
                data_in = _encrypted_data.strip().encode()  # type: str
                _hmac = hmac.new(key_mac, data_in, hashlib.sha256)
                _hmaced = _hmac.digest()
                local_signature = base64.b64encode(_hmaced).decode()
                msg_signature = data['signature']['hmac']

                # exit if the HMAC is incorrect
                if local_signature != msg_signature:
                    logger.info("Message signature is invalid")
                    return None

                # get the IV from the message
                raw_iv = data['signature']['iv']
                iv = base64.b64decode(raw_iv.encode())

                _encrypted_data = base64.b64decode(data_in)
                # create an encryptor and encrypt the data
                aes = Decrypter(AESModeOfOperationCBC(key_enc, iv=iv))
                data_out = aes.feed(_encrypted_data)
                data_out += aes.feed()  # type: bytes

                data['payload'] = json.loads(data_out.decode())
                data['signature'] = None

                return data
            except Exception as ex:
                logger.error("Exception when decrypting data", cause=str(ex))
                return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
 This file is owned exclusively by Smart Arcs Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
"""
import json
import os
import socket
import sys
import time
import traceback
# noinspection PyProtectedMember
from ipaddress import ip_network, ip_address, _BaseAddress, _BaseNetwork
from multiprocessing import current_process, Event

import dns.zone

import logbook
import requests

from keychest_agent.config import KCConfig
from keychest_agent.logger import logger, TCPHandler
from keychest_agent.misc import TlsUtils, DNSUtils, ControlUtils
from keychest_agent.proxy import AgentProxy
from keychest_agent.threadpool import DiscoveryPool, test_ip
from keychest_agent.security import Security

__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'


class Agent(object):

    def __init__(self, args):
        self.banned_space = []
        self.permitted_space = []  # list of pairs
        self.counter = 0
        self.config = None
        self.last_result = 0
        self.args = args
        self.config_ready = False
        self.log_dispatcher_thread = None
        self.logger_stop_event = Event()
        self.active_proxies = {}
        self.host_inventory = {}
        self.proxy_ports = []
        self.init_config()
        self.update_proxy_range()

    def update_proxy_range(self):
        proxy_range = KCConfig.config.agent_proxy_ports
        if isinstance(proxy_range, list):
            self.proxy_ports = []
            for _port in proxy_range:
                if isinstance(_port, int) and (_port > 1024) and (_port < 65536):
                    self.proxy_ports.append(_port)
        else:
            proxy_range = str(proxy_range)
            _split = proxy_range.split('-')
            if len(_split) == 1:
                self.proxy_ports = [int(proxy_range)]
            else:
                self.proxy_ports = list(range(int(_split[0]), int(_split[1]) + 1))

    def run(self):
        pass

    def restart_itself(self):
        # https://github.com/cherrypy/cherrypy/blob/0857fa81eb0ab647c7b59a019338bab057f7748b/
        # cherrypy/process/wspbus.py#L305
        # args = sys.argv[:]
        # self.log('Re-spawning %s' % ' '.join(args))
        #
        # args.insert(0, sys.executable)
        # if sys.platform == 'win32':
        #     args = ['"%s"' % arg for arg in args]
        #
        # os.chdir(_startup_cwd)
        # os.execv(sys.executable, args)
        pass

    def return_code(self, code=0):
        self.last_result = code
        return code

    def init_config(self):
        """
        Initializes configuration
        :return:
        """
        self.config = KCConfig.read_configuration()
        if self.config is None or not self.config.has_nonempty_config():
            sys.stderr.write('Configuration is empty: %s\nCreating a new one ... \n  (review and update '
                             'authentication credentials)\n' % KCConfig.get_config_file_path())
            self.config = KCConfig.default_config()
            KCConfig.write_configuration(self.config)

        self.config_ready = True

        try:
            if not os.path.exists(KCConfig.config.logdir):
                os.makedirs(KCConfig.config.logdir, mode=0o755)
            if not os.path.exists(KCConfig.config.piddir):
                os.makedirs(KCConfig.config.piddir, mode=0o755)
            if not os.path.exists(KCConfig.config.sockdir):
                os.makedirs(KCConfig.config.sockdir, mode=0o755)
            if not os.path.exists(KCConfig.config.rundir):
                os.makedirs(KCConfig.config.rundir, mode=0o755)
        except Exception as ex:
            logger.error("Can't create a config folder", cause=str(ex))

        # clean pid folder contents
        for the_file in os.listdir(KCConfig.config.piddir):
            file_path = os.path.join(KCConfig.config.piddir, the_file)
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
            except Exception as ex:
                logger.error("Can't clean pid folder", cause=str(ex))
                pass
        # let's process IP addresses into an operational form
        self.set_permitted_space()
        self.set_banned_space()
        # also update_proxy_range ....

    def set_permitted_space(self):
        permitted = KCConfig.config.permitted_space
        _step1 = permitted.split(',')

        for _range in _step1:
            # noinspection PyBroadException
            try:
                _range = _range.strip()
                _parts = _range.split(':')
                _is_range = len(_parts[0].split('/')) > 1
                if len(_parts) > 1:
                    _port = int(_parts[1])
                else:
                    _port = 0
                if _is_range:
                    _item = (ip_network(_parts[0]), _port)
                else:
                    _item = (ip_address(_parts[0]), _port)
                self.permitted_space.append(_item)
            except Exception:
                logger.error("The definition of permitted space is incorrect", query=_range)

    def set_banned_space(self):
        banned = KCConfig.config.banned_space
        _step1 = banned.split(',')

        for _range in _step1:
            if len(_range.strip()) == 0:
                continue
            # noinspection PyBroadException
            try:
                _range = _range.strip()
                _parts = _range.split(':')
                _is_range = len(_parts[0].split('/')) > 1
                if len(_parts) > 1:
                    _port = int(_parts[1])
                else:
                    _port = 0
                if _is_range:
                    _item = (ip_network(_parts[0]), _port)
                else:
                    _item = (ip_address(_parts[0]), _port)
                self.banned_space.append(_item)
            except Exception:
                logger.error("The definition of banned space is incorrect", query=_range)

    # noinspection PyUnusedLocal
    @staticmethod
    def setup_logging(level=logbook.DEBUG, true_socket=False):

        if true_socket:
            local_sock = socket.gethostbyname("localhost")
            port = KCConfig.config.internal_socket
            _handler = TCPHandler(None, flush_threshold=1, flush_time=1, ip=local_sock, port=port,
                                  extra_fields={'process': current_process().name})
        else:
            sock_file = os.path.join(os.path.join(KCConfig.config.sockdir, 'logbook.sock'))
            _handler = TCPHandler(sock_file, flush_threshold=1, flush_time=1,
                                  extra_fields={'process': current_process().name})
        # noinspection PyBroadException
        try:
            _handler.pop_application()
        except:
            pass
        _handler.push_application()

    # noinspection PyMethodMayBeStatic
    def do_registration(self):
        """
        We need to register after installation - this is the code
        :return:
        """

        logger.debug("Starting registration request")
        registration = {
            "command": "registration",
            "signature": None,
            "payload": {
                "config": KCConfig.create_config_dict(KCConfig.config.json['config'])
            }
        }
        result = False
        headers = {'Content-type': 'application/json', 'Accept': 'application/json'}
        complete_url = ""
        try:
            # Send data
            # add  the agent_id to the URL
            complete_url = KCConfig.config.control_server
            if complete_url[len(complete_url)-1] != "/":
                complete_url += "/" + KCConfig.config.agent_email
            else:
                complete_url += KCConfig.config.agent_email

            encrypted_registration = Security.secure(registration, KCConfig.config.agent_apikey)
            data_text = json.dumps(encrypted_registration)
            r = requests.post(url=complete_url, data=data_text, headers=headers, timeout=KCConfig.config.rest_timeout)
            reg_res = r.json()
            resp_code = r.status_code
            if resp_code != 200:
                logger.warning("Registration response signals an error", code=resp_code,
                               response=reg_res)
                return result
            # let's process the registration result
            if 'status' in reg_res:
                if (reg_res['status'] == 'success') and ('data' in reg_res) \
                        and ('email' in reg_res['data']) and ('apikey' in reg_res['data']):
                    # all good, let's do it
                    apikey = reg_res['data']['apikey']
                    email = reg_res['data']['email']
                    name = reg_res['data']['name']
                    KCConfig.config.agent_email = email
                    KCConfig.config.agent_apikey = apikey
                    KCConfig.config.agent_name = name
                    # stop so we can start sending logs with the new
                    time.sleep(8)  # log dispatcher will pickup the new email in less than 5 seconds
                    try:
                        KCConfig.write_configuration(KCConfig.config)
                        KCConfig.read_configuration(force=True)
                        result = True
                    except Exception as ex:
                        logger.error("Failed to store registration data", cause=str(ex),
                                     request=KCConfig.config.to_string())
                else:
                    logger.error("Registration failed", request=data_text,
                                 response=json.dumps(reg_res))
        except Exception as ex:
            logger.error("Error connecting to control channel", url=complete_url, cause=str(ex))
        finally:
            logger.debug("Registration completed")
            return result

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def do_ping(self, do_discovery):
        """
        We regularly check for updates in the discovery configuration
        :return:
        """

        logger.debug("Starting a regular PING request with configuration check")
        discovery = {
            "command": "ping",
            "signature": None,
            "payload": {
                "config": KCConfig.create_config_dict(KCConfig.config.json['config'])
            }
        }
        result = False
        headers = {'Content-type': 'application/json', 'Accept': 'application/json'}
        complete_url = ""
        try:
            # Send data
            # add  the agent_id to the URL
            complete_url = KCConfig.config.control_server
            if complete_url[len(complete_url)-1] != "/":
                complete_url += "/" + KCConfig.config.agent_email
            else:
                complete_url += KCConfig.config.agent_email

            encrypted_discovery = Security.secure(discovery, KCConfig.config.agent_apikey)
            data_text = json.dumps(encrypted_discovery)
            r = requests.post(url=complete_url, data=data_text, headers=headers, timeout=KCConfig.config.rest_timeout)
            reg_res = r.json()
            resp_code = r.status_code
            if resp_code != 200:
                return result

            # let's process the registration result
            if 'status' in reg_res:
                if (reg_res['status'] == 'success') and ('data' in reg_res):
                    # all good, let's do it - first we may want to check the ip addresses are allowed
                    data = reg_res['data']
                    resolved = []
                    try:
                        if 'config' in data:
                            # we are getting a new config file, let's try to update it
                            if KCConfig.config.local_config:
                                logger.info("Keychest service pushed configuration changes but we manage the configuration locall")
                            else:
                                try:
                                    new_config = KCConfig.create_config_dict(data['config'])
                                    # test if it is for us
                                    if ('agent_email' not in new_config) or new_config['agent_email'] != KCConfig.config.agent_email:
                                        _name = new_config['agent_email'] if 'agent_email' in new_config else "<undefined>"
                                        logger.error("Received configuration for another agent", name=_name)
                                        return  # we should probably stop processing the rest
                                    identical = True
                                    for key, value in new_config.items():
                                        if key in ['git_version', 'agent_email', 'control_server'] or KCConfig.is_local_item(key):
                                            continue
                                        local_value = KCConfig.config.get_config(key)
                                        if local_value != value:
                                            try:
                                                KCConfig.config.set_config(key, value)
                                                new_value = KCConfig.config.get_config(key)
                                                if new_value != value:
                                                    logger.error("Failed update configuration value",
                                                                 name=key)
                                                else:
                                                    identical = False
                                                    logger.info("Configuration value has been updated",
                                                                name=key, old_value=local_value, new_value=new_value)
                                            except Exception as ex:
                                                logger.error("Failed to udpate configuration value - exception",
                                                             name=key, cause=str(ex))
                                    if not identical:
                                        KCConfig.write_configuration(KCConfig.config)
                                    self.set_banned_space()
                                    self.set_permitted_space()
                                except Exception as ex:
                                    logger.error("Error parsing new configuration", cause=str(ex))
                                pass
                            pass

                        # first we process DNS if any is provided -> it will replace IP range scanning
                        if ('dns_file' in data) and (data['dns_file'] is not None):
                            # try to read dns file
                            # the entry can have several paths
                            current_folder = os.getcwd()
                            for each_path in data['dns_file']:
                                # we  get an object
                                filename =  each_path['dns_file_path'] if 'dns_file_path' in each_path else None
                                if filename is None:
                                    logger.error("DNS file path not present", query=each_path)
                                    continue

                                # read the domain if present and normalize - it may be needed if not defined in the file
                                main_domain = each_path['dns_file_domain'] \
                                    if ('dns_file_domain' in each_path) and (each_path['dns_file_domain'] is not None)\
                                    else ""  # type: str
                                if (main_domain is not None) and (len(main_domain.strip()) < 1):
                                    main_domain =  ""
                                if main_domain is not None:
                                    main_domain = main_domain.strip().lower()

                                if os.path.exists(filename):
                                    if os.path.isfile(filename):
                                        try:
                                            # we may not have sufficient privilege to read
                                            open(filename, 'r').close()
                                            if len(main_domain):
                                                zone = dns.zone.from_file(filename, origin=main_domain)
                                            else:
                                                zone = dns.zone.from_file(filename)
                                            # update  origin / main domain
                                            if len(str(zone.origin)) > 1:
                                                main_domain = str(zone.origin).lower()
                                                if main_domain[len(main_domain)-1] == '.':
                                                    main_domain = main_domain[:-1]  # remove trailing dot

                                            name_map = DNSUtils.process_zone(zone, main_domain)
                                            filtered = ControlUtils.filter_banned(self, name_map)
                                            for each_filtered in filtered:
                                                if each_filtered not in resolved:
                                                    resolved.append(each_filtered)

                                        except dns.exception.DNSException as ex:
                                            logger.error("DNS zone file doesn't contain its "
                                                         "origin domain or other DNS-related error",
                                                         query=each_path, cause=str(ex))
                                        except Exception as ex:
                                            logger.error("DNS zone file found but can't be read", cause=str(ex),
                                                         query=str(filename), cwd=current_folder)
                                    else:
                                        logger.error("DNS file path does not point to a regular file",
                                                     query=str(filename), cwd=current_folder)
                                else:
                                    logger.error("DNS file doesn't exist or it can't be read",
                                                 query=str(filename), cwd=current_folder)
                            logger.info("DNS zone file processing completed", count=len(resolved))
                            pass
                        if ('dns_transfer' in data) and (data['dns_transfer'] is not None):  # type: dict
                            # try to do zone transfer
                            dnshost = None
                            dnsdomain = None
                            dnsip =  None
                            # noinspection PyTypeChecker
                            if 'dns_transfer_addr' in data['dns_transfer']:
                                # noinspection PyTypeChecker
                                dnshost = data['dns_transfer']['dns_transfer_addr']
                            # noinspection PyTypeChecker
                            if 'dns_transfer_domain' in data['dns_transfer']:
                                # noinspection PyTypeChecker
                                dnsdomain = data['dns_transfer']['dns_transfer_domain']
                            if dnshost is None:
                                logger.error("Request for DNS zone transfer has incorrect format - dnshost is missing")
                            if dnsdomain is None:
                                logger.error("Request for DNS zone transfer has incorrect format - dnsdomain is missing")

                            # let's check the host name first - we will run DNS queries against it
                            try:
                                dnsip = socket.gethostbyname(dnshost)
                            except Exception as ex:
                                logger.error("DNS server name provided is invalid", query=dnshost, cause=str(ex))

                            if dnshost and dnsdomain and dnsip:
                                # let's try the zone transfer
                                try:
                                    zone = dns.zone.from_xfr(dns.query.xfr(dnsip, dnsdomain))
                                    logger.info("DNS zone transfer completed")
                                    if len(str(zone.origin)) > 1:
                                        dnsdomain = str(zone.origin).lower()
                                        if dnsdomain[len(dnsdomain) - 1] == '.':
                                            dnsdomain = dnsdomain[:-1]  # remove trailing dot

                                    name_map = DNSUtils.process_zone(zone, dnsdomain)
                                    filtered = ControlUtils.filter_banned(self, name_map)
                                    for each_filtered in filtered:
                                        if each_filtered not in resolved:
                                            resolved.append(each_filtered)
                                    logger.info("DNS transfer completed and processed",
                                                count=len(filtered), total=len(resolved))
                                except Exception as ex:
                                    logger.error("DNS zone transfer request failed", host=dnsip,
                                                 domain=dnsdomain, cause=str(ex))

                            # at this point, name_map has a processed DNS zone records - let's copy it over to a
                            # result array while applying restrictions
                            #
                            # #  and let's send the results back
                            # network = {
                            #     "command": "network",
                            #     "signature": None,
                            #     "payload": {
                            #         "target": resolved
                            #     }
                            # }
                            # headers = {'Content-type': 'application/json'}
                            # try:
                            #     # Send data
                            #     # add  the agent_id to the URL
                            #     complete_url = os.path.join(KCConfig.config.control_server,
                            #                                 KCConfig.config.agent_email)
                            #     data_text = json.dumps(network)
                            #     r = requests.post(url=complete_url, data=data_text, headers=headers)
                            #     reg_res = r.json()
                            # except Exception as ex:
                            #     logger.error("Failed to send network nodes", response=reg_res, cause=str(ex))

                            pass
                        if (('ip_range' in data) and (len(data['ip_range']) > 0)) or (len(resolved) > 0):
                            resolved_done = []
                            # option 1 -> no DNS resolved, we will do a brute IP range test
                            if ('ip_range' in data) and (resolved == []):
                                # iterate over data items
                                pool = DiscoveryPool(KCConfig.config.discovery_threads)
                                time_now = time.time()
                                for each_range in data['ip_range']:
                                    if 'network' in each_range:
                                        port = each_range['port']
                                        # noinspection PyBroadException
                                        try:
                                            network = ip_network(each_range['network'], False)
                                            pass
                                        except Exception:
                                            continue
                                        for _each_ip in list(network):
                                            # first checked, if this IP address has been processed
                                            proceed = False
                                            if str(_each_ip) in self.host_inventory:
                                                if (time_now - self.host_inventory[str(_each_ip)]) > KCConfig.config.discovery_schedule:
                                                    proceed = True
                                                    logger.debug("Existing IP will be re-tested",
                                                                 ip=str(_each_ip), time=self.host_inventory[str(_each_ip)])
                                                else:
                                                    logger.trace("The IP address has already been tested",
                                                                 ip=str(_each_ip), time=self.host_inventory[str(_each_ip)])
                                            else:
                                                logger.info("New IP address", ip=str(_each_ip))
                                                proceed = True
                                            # then we can do name resolution
                                            if proceed:
                                                # noinspection PyBroadException,PyUnusedLocal
                                                pool.enqueue(test_ip, _each_ip, port, KCConfig.config.connect_timeout)
                                            self.host_inventory[str(_each_ip)] = time_now
                                # work is ready, let's start processing
                                pool.run()
                                time.sleep(2)
                                while not pool.done():
                                    _r = pool.results()
                                    if isinstance(_r, list):
                                        resolved_done += _r
                                    time.sleep(5)
                                # and a final pull
                                _r = pool.results()
                                if isinstance(_r, list):
                                    resolved_done += _r

                                pool.abort(block=True)
                                del pool  # close the pool and all its threads
                                pass
                            elif 'ip_range' in data:  # we did some DNS zone processing -> we use it as a filter
                                _networks = []
                                if 'ip_range' in data:
                                    for each_range in data['ip_range']:
                                        if 'network' in each_range:
                                            port = each_range['port']
                                            # noinspection PyBroadException
                                            try:
                                                network = ip_network(each_range['network'], False)
                                                _networks.append({'net': network, 'port': port})
                                            except Exception:
                                                continue
                                    # test DNS zone items and give them ports
                                    for each_host in resolved:
                                        try:
                                            _ip = ip_address(each_host['ip'])
                                            for each_network in _networks:
                                                if _ip in each_network['net']:
                                                    resolved_done.append({
                                                        'name': each_host['name'],
                                                        'ip': each_host['ip'],
                                                        'port': each_network['port']
                                                    })
                                        except Exception as ex:
                                            logger.warning("Failed to parse/process IP address",
                                                           query=each_host, cause=str(ex))
                                            pass
                            else:  # no IP range - we take everything with port 443
                                for each_host in resolved:
                                    resolved_done.append({
                                        'name': each_host['name'],
                                        'ip': each_host['ip'],
                                        'port': KCConfig.config.default_port
                                    })
                                pass
                            # now we need to send data back
                            network = {
                                "command": "network",
                                "signature": None,
                                "payload": {
                                    "target": resolved_done
                                }
                            }
                            headers = {'Content-type': 'application/json', 'Accept': 'application/json'}
                            try:
                                # Send data
                                # add  the agent_id to the URL
                                complete_url = KCConfig.config.control_server
                                if complete_url[len(complete_url) - 1] != "/":
                                    complete_url += "/" + KCConfig.config.agent_email
                                else:
                                    complete_url += KCConfig.config.agent_email

                                encrypted_network = Security.secure(network, KCConfig.config.agent_apikey)
                                data_text = json.dumps(network)
                                r = requests.post(url=complete_url, data=data_text,
                                                  headers=headers, timeout=KCConfig.config.rest_timeout)
                                reg_res = r.json()
                                resp_code = r.status_code
                                if resp_code != 200:
                                    logger.warning("Error when sending a list of resolved addresses to Keychest",
                                                   code=resp_code, response=reg_res)
                                    result = False
                                    return result
                            except Exception as ex:
                                logger.error("Failed to send network nodes", response=reg_res, cause=str(ex))
                                result = False
                        if 'ldap_pull' in data:
                            # try to pull certificate data from ldap
                            pass
                        if 'ad_pull' in data:
                            # try to pull data from AD
                            # noinspection PyTypeChecker
                            if (KCConfig.config.ad_password is None) or (KCConfig.config.ad_user is None):
                                logger.error("Request for AD access without username or password in local config file")
                            elif 'adhost' not in data['ad_pull']:
                                logger.error("Request for AD access does not contain the host name")
                            else:
                                adhost = None
                                addomain = None
                                # noinspection PyTypeChecker
                                if 'adhost' in data['ad_pull']:
                                    # noinspection PyUnusedLocal
                                    adhost = data['adhost']
                                # noinspection PyTypeChecker
                                if 'addomain' in data['ad_pull']:
                                    # noinspection PyUnusedLocal
                                    addomain = data['addomain']
                            pass
                        if 'pem_file' in data:
                            # try to read certificates
                            pass
                        result = True
                    except Exception as ex:
                        tb = "; ".join(traceback.format_exc(3).splitlines())
                        logger.error("Failed to process data for discovery", cause=str(ex),
                                     traceback=tb, request=reg_res)
                        result = False
                else:
                    logger.error("Discovery request failed", request=data_text,
                                 response=json.dumps(reg_res))
                    result = False
            else:
                logger.error("Discovery request failed - structure", request=data_text,
                             response=json.dumps(reg_res))
                result = False
        except Exception as ex:
            logger.error("Error connecting to control channel", url=complete_url, ccause=str(ex))
            return result
        finally:
            logger.debug("A regular PING completed")
            return result

    def do_request_jobs(self):
        """
        Send a request for new jobs and return a dictionary keyed by the proxy ports
        :return:
        :rtype: dict
        """

        logger.debug("Requesting new jobs from KeyChest service")
        # get rid of active ports
        threads_stopped = []
        result = None
        for key, value in self.active_proxies.items():
            if not value.is_alive():
                threads_stopped.append(key)
                if key not in self.proxy_ports:
                    self.proxy_ports.append(key)  # return the port to the pool of available

        for _thread in threads_stopped:
            del self.active_proxies[_thread]

        # we should now have a list of free ports

        # create info about the agent
        free_proxies = {
            "command": "jobs",
            "signature": "sign",
            "payload": {
                "ports": self.proxy_ports,
                "reverse": KCConfig.config.reverse_proxy,
                "timestamp": int(time.time()),
                "counter": self.counter
            }
        }
        self.counter += 1
        headers = {'Content-type': 'application/json', 'Accept': 'application/json'}
        complete_url = ""
        try:
            # Send data
            # add  the agent_id to the URL
            complete_url = KCConfig.config.control_server
            if complete_url[len(complete_url)-1] != "/":
                complete_url += "/" + KCConfig.config.agent_email
            else:
                complete_url += KCConfig.config.agent_email

            encrypted_free_proxies = Security.secure(free_proxies, KCConfig.config.agent_apikey)
            data_text = json.dumps(encrypted_free_proxies)
            r = requests.post(url=complete_url, data=data_text, headers=headers, timeout=KCConfig.config.rest_timeout * 2)
            jobs = r.json()
            resp_code = r.status_code
            if resp_code != 200:
                logger.warning("Request for new jobs failed", code=resp_code,
                               response=jobs)
                result = None
                return result
            if ('status' in jobs) and (jobs['status'] == 'success'):
                if 'data' in jobs:
                    result = jobs['data']
                else:
                    logger.warning("Control channel didn't return 'data' item in its response",
                                   response=json.dumps(jobs))
                    result = {}
            else:
                logger.error("Control channel request failed", request=data_text,
                             response=json.dumps(jobs))
                result = None  # signal an error
        except Exception as ex:
            logger.error("Error connecting to control channel", url=complete_url, ccause=str(ex))
        finally:
            logger.debug("Job request completed")
            return result

    # noinspection PyMethodMayBeStatic
    def is_target_allowed(self, target, target_port, permitted_space=None, banned_space=None):
        """
        Checks if the target is in the allowed IP and port range
        :param target:
        :type target: str
        :param target_port:
        :type target_port: int
        :param permitted_space:
        :type permitted_space: list
        :param banned_space:
        :type banned_space: list
        :return:
        :rtype: bool
        """

        if ((permitted_space is None) or len(permitted_space) == 0)\
                and ((banned_space is None) or (len(banned_space) == 0)):
            return True

        # noinspection PyBroadException
        try:
            if TlsUtils.is_ip(target):
                _target = target
            else:
                _target = socket.gethostbyname(target)
            # we have an IP address

            for (banned_ip, banned_port) in banned_space:
                if (banned_port > 0) and (banned_port == target_port):
                    if isinstance(banned_ip, _BaseAddress):
                        if ip_address(target) == banned_ip:
                            return False  # banned
                    elif isinstance(banned_ip, _BaseNetwork):
                        if ip_address(target) in banned_ip:
                            return False  # banned

            # now let's check if permitted
            if len(permitted_space) == 0:
                return True
            else:
                permitted = False
                for (permitted_ip, permitted_port) in permitted_space:
                    if (permitted_port == 0) or (permitted_port == target_port):
                        if isinstance(permitted_ip, _BaseAddress):
                            if ip_address(target) == permitted_ip:
                                return True  # banned
                        elif isinstance(permitted_ip, _BaseNetwork):
                            if ip_address(target) in permitted_ip:
                                return False  # banned
                return permitted
        except Exception:
            return False

    # noinspection PyMethodMayBeStatic
    def start_new_proxy(self, port, job, permitted_space=None, banned_space=None):
        """
        This method will launch a new proxy to execute an audit job
        :param banned_space:
        :param permitted_space:
        :param port: for the the upstream proxy - it's from a pool, i.e. it limits number of threads
        :type port: int
        :param job:
        :return:
        :rtype: Union[Thread,None]
        """

        if job['type'] == 'PeriodicIpScanJob':
            target_port = job['port'] if 'port' in job else 443
            proto = job['proto'] if 'proto' in job else 'tcp'
            ip_int_beg = TlsUtils.ip_to_int(job['target']['ip_beg'])
            ip_int_end = TlsUtils.ip_to_int(job['target']['ip_end'])
            target = list(TlsUtils.iter_ips(ip_start_int=ip_int_beg, ip_stop_int=ip_int_end))

            upstream_port = job['upstream_port'] if 'upstream_port' in job else KCConfig.config.audit_server_port
            upstream_ip = job['upstream_host'] if 'upstream_host' in job else KCConfig.config.upstream_host
            upstream = (upstream_ip, upstream_port)
            job_id = job['key']
            send_upstream = json.dumps(job).encode()
            destination = (target, target_port)
            logger.debug("Starting a new proxy to audit a target - PeriodicIpScanJob", target=upstream, id=job_id)
            new_proxy = AgentProxy(proto, upstream, destination, job_id, send_upstream, reverse=True)
            # new_proxy.run()
            new_proxy.start()
            return new_proxy
        elif job['type'] == 'PeriodicObjectJob':
            target_port = job['port'] if 'port' in job else 443
            proto = job['proto'] if 'proto' in job else 'tcp'
            target = job['ip']
            if self.is_target_allowed(target, target_port, permitted_space, banned_space):
                upstream_port = job['upstream_port'] if 'upstream_port' in job else KCConfig.config.audit_server_port
                upstream_ip = job['upstream_ip'] if 'upstream_ip' in job else KCConfig.config.upstream_host
                upstream = (upstream_ip, upstream_port)
                job_id = job['key']
                send_upstream = json.dumps(job).encode()
                destination = (target, target_port)
                logger.debug("Starting a new proxy to audit a target - PeriodicObjectJob", target=upstream, id=job_id)
                new_proxy = AgentProxy(proto, upstream, destination, job_id, send_upstream, reverse=True)
                # new_proxy.run()
                new_proxy.start()
                return new_proxy
            else:
                logger.info("Request for banned IP address", query="%s:%d" % (target, target_port))
                return None
        else:
            return None

    def work_loop(self):

        last_discovery = 0
        last_ping = 0
        last_job_request = 0
        registered = False

        exit_flag = Event()
        while not (registered and KCConfig.config.do_register):
            try:
                # we start a RESTful client = control channel
                if KCConfig.config.agent_email != "dummy@keychest.net":
                    registered = True
                else:
                    registered = False
                while not (registered and KCConfig.config.do_register):
                    if not registered:
                        registered = self.do_registration()
                    else:  # we have a valid email and apikey - let's get cracking
                        registered = True
                        time_now = time.time()
                        if (time_now - last_ping) > KCConfig.config.control_schedule:
                            discovery_interval = (time_now - last_discovery) > KCConfig.config.discovery_schedule
                            if not KCConfig.config.reverse_proxy:
                                logger.error("This agent only supports the proxy mode 'reverse' set to True")
                                discovery_interval = False
                            success = self.do_ping(discovery_interval)
                            logger.debug("Ping completed", result=success)
                            if success:
                                last_ping = time_now
                                if discovery_interval:
                                    logger.debug("Discovery completed", result=success)
                                    last_discovery = time_now

                        # only do the following if in the reverse mode
                        if ((time_now - last_job_request) > KCConfig.config.job_schedule) \
                                and KCConfig.config.reverse_proxy:
                            self.update_proxy_range()
                            jobs = self.do_request_jobs()
                            if isinstance(jobs, dict) and len(jobs) > 0:
                                last_job_request = time.time()
                                for key, value in jobs.items():
                                    try:
                                        new_proxy = self.start_new_proxy(int(key), value)
                                        if new_proxy:
                                            self.active_proxies[int(key)] = new_proxy
                                    except Exception as ex:
                                        logger.error("Can't start a new proxy", cause=str(ex), key=key)
                            elif isinstance(jobs, dict):
                                last_job_request = time.time()  # no job, just update timestamp
                            else:
                                # job request failed
                                pass
                    # wait for next tick - wait smart
                    time_now = time.time()
                    wait1 = KCConfig.config.control_schedule - (time_now - last_ping)
                    wait2 = KCConfig.config.job_schedule - (time_now - last_job_request)
                    if wait1 < wait2:
                        logger.debug("Sleeping", delay=wait1)
                        exit_flag.wait(wait1)
                    else:
                        logger.debug("Sleeping", delay=wait2)
                        exit_flag.wait(wait2)
                pass
                # else:
                #     # we start a RESTful server listening for control channel requests
                #     pass
            except Exception as ex:
                tb = "; ".join(traceback.format_exc(5).splitlines())
                logger.error("Exception in the Agent work_loop", cause=str(ex), traceback=tb)

        # if the user requested registration only - signal it out
        return registered and KCConfig.config.do_register

    # noinspection PyMethodMayBeStatic
    def is_banned(self, _each_ip, port=443):
        each_ip = ip_address(_each_ip)
        try:
            for each_net in self.banned_space:
                if isinstance(each_net[0], _BaseAddress):
                    if (each_ip == each_net[0]) and ((port == each_net[1]) or (each_net[1] == 0)):
                        return True
                else:  # it's a network
                    if (each_ip in each_net[0]) and ((port == each_net[1]) or (each_net[1] == 0)):
                        return True

            if len(self.permitted_space) == 0:  # by default, we assume there are no restrictions
                return False
            for each_net in self.permitted_space:
                if isinstance(each_net[0], _BaseAddress):
                    if (each_ip == each_net[0]) and ((port == each_net[1]) or (each_net[1] == 0)):
                        return False
                else:  # it's a network
                    if (each_ip in each_net[0]) and ((port == each_net[1]) or (each_net[1] == 0)):
                        return False

            return True  # not allowed -> banned
        except Exception as ex:
            tb = "; ".join(traceback.format_exc(5).splitlines())
            logger.error("Exception when testing for banned IP addresses",
                         cause=str(ex), traceback=tb)
        return True   # this is the default



from typing import Union

import win32serviceutil
import win32service
import win32event
import servicemanager
import socket

from multiprocessing import Process, freeze_support

import sys

from keychest_agent import main

register = False

class WinMain (win32serviceutil.ServiceFramework):
    _svc_name_ = "KeyChest"
    _svc_display_name_ = "KeyChest"

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, *args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)
        self.stop_requested = False
        socket.setdefaulttimeout(120)
        self.body = None    # type: Union[None,Process]

    # noinspection PyPep8Naming
    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)
        self.body.terminate()
        # pid = self.body.pid
        # os.kill(pid, signal.SIGTERM)
        # self.body.stop()
        self.ReportServiceStatus(win32service.SERVICE_STOPPED)
        self.stop_requested = True

    # noinspection PyPep8Naming
    def SvcDoRun(self):

        global register

        self.ReportServiceStatus(win32service.SERVICE_START_PENDING)
        if register:
            self.ReportServiceStatus(win32service.SERVICE_RUNNING)
        else:
            self.body = Process(target=main.main, args=())
            self.body.daemon = False
            self.ReportServiceStatus(win32service.SERVICE_RUNNING)
            self.body.start()
            servicemanager.LogMsg(
                servicemanager.EVENTLOG_INFORMATION_TYPE,
                servicemanager.PYS_SERVICE_STARTED,
                (self._svc_name_, '')
            )

            # self.timeout = 5000  # frequency seconds (timeout is in milliseconds)
            # rc = None
            win32event.WaitForSingleObject(self.hWaitStop, win32event.INFINITE)

            # while rc != win32event.WAIT_OBJECT_0:
            #     # Wait for stop signal, loop on timeout
            #     rc = win32event.WaitForSingleObject(self.hWaitStop, self.timeout)


if __name__ == '__main__':
    freeze_support()
    if (len(sys.argv) > 1) and ((sys.argv[1] == '--register') or (sys.argv[1] == '--run') or (sys.argv[1] == '--getid')):
        if sys.argv[1] == '--register':
            main.main()
        else:
            main.main()
        sys.exit(0)
    elif len(sys.argv) == 1:
        servicemanager.Initialize()
        servicemanager.PrepareToHostSingle(WinMain)
        servicemanager.StartServiceCtrlDispatcher()
    else:
        win32serviceutil.HandleCommandLine(WinMain)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) Smart Arcs Ltd, registered in the United Kingdom.
 This file is owned exclusively by Smart Arcs Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@keychest.net>, 2019
"""
__copyright__ = 'Smart Arcs Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'

# this is mostly from:
# http://code.activestate.com/recipes/577187-python-thread-pool/

from threading import Thread, Event
from time import sleep
import socket
from multiprocessing import JoinableQueue

from keychest_agent.logger import logger


# default exception handler. if you want to take some action on failed tasks
# maybe add the task back into the queue, then make your own handler and pass it in
def default_handler(name, exception, *args, **kwargs):
    logger.error("Discovery worker error", name=name, cause=str(exception),
                 args=repr(args), kwargs=repr(kwargs))
    pass


def test_ip(_each_ip, port, timeout=5):
    # noinspection PyBroadException
    try:
        each_ip = str(_each_ip)
        _dns = False
        name = None
        # noinspection PyBroadException
        try:
            dns_resolv = socket.gethostbyaddr(each_ip)
            _dns = True
            name = dns_resolv[0]
        except:
            _dns = False

        _open = False
        # noinspection PyBroadException
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            sock.settimeout(timeout)
            _result = sock.connect_ex((each_ip, port))
            if _result == 0:
                _open = True
            sock.close()
        except:
            pass
        logger.debug("Tested a new IP address", ip=each_ip, port=port, result=_open, name=name)
        if _open:
            return {'name': name, 'ip': each_ip, 'port': port}
        else:
            return None
    except Exception as ex:
        logger.trace("DNS resolve failed", cause=str(ex))
        # if dns doesn't work it will return an exception
        return None
        pass

# class for workers
class Worker(Thread):
    """Thread executing tasks from a given tasks queue"""

    def __init__(self, name, queue, results, abort, idle, exception_handler):
        Thread.__init__(self)
        self.name = name
        self.queue = queue
        self.results = results
        self.abort = abort
        self.idle = idle
        self.exception_handler = exception_handler
        self.daemon = True
        self.start()

    """Thread work loop calling the function with the params"""

    def run(self):
        # keep running until told to abort
        while not self.abort.is_set():
            # noinspection PyBroadException
            try:
                # get a task and raise immediately if none available
                func, args, kwargs = self.queue.get(False)
                self.idle.clear()
            except:
                # no work to do
                # if not self.idle.is_set():
                #  print >> stdout, '%s is idle' % self.name
                self.idle.set()
                continue

            try:
                # the function may raise
                _result = func(*args, **kwargs)
                if _result is not None:
                    self.results.put(_result)
            except Exception as e:
                # so we move on and handle it in whatever way the caller wanted
                self.exception_handler(self.name, e, args, kwargs)
            finally:
                # task complete no matter what happened
                self.queue.task_done()


# class for thread pool
class DiscoveryPool:
    """Pool of threads consuming tasks from a queue"""

    def __init__(self, thread_count, batch_mode=False, exception_handler=default_handler):
        # batch mode means block when adding tasks if no threads available to process
        self.queue = JoinableQueue(thread_count if batch_mode else 0)
        self.resultQueue = JoinableQueue(0)
        self.thread_count = thread_count
        self.exception_handler = exception_handler
        self.aborts = []
        self.idles = []
        self.threads = []

    def __del__(self):
        """
        Tell my threads to quit

        :return:
        """
        self.abort()

    def run(self, block=False):
        """
        Start the threads, or restart them if you've aborted
        :param block:
        :return:
        """
        # either wait for them to finish or return false if some arent
        if block:
            while self.alive():
                sleep(1)
        elif self.alive():
            return False

        # go start them
        self.aborts = []
        self.idles = []
        self.threads = []
        for n in range(self.thread_count):
            abort = Event()
            idle = Event()
            self.aborts.append(abort)
            self.idles.append(idle)
            thread = Worker('poolthread-%d' % n, self.queue, self.resultQueue, abort, idle, self.exception_handler)
            self.threads.append(thread)
        return True

    """Add a task to the queue"""

    def enqueue(self, func, *args, **kargs):
        self.queue.put((func, args, kargs))

    """Wait for completion of all the tasks in the queue"""

    def join(self):
        self.queue.join()

    """Tell each worker that its done working"""

    def abort(self, block=False):
        # tell the threads to stop after they are done with what they are currently doing
        for a in self.aborts:
            a.set()
        # wait for them to finish if requested
        while block and self.alive():
            sleep(1)

    """Returns True if any threads are currently running"""

    def alive(self):
        return True in [t.is_alive() for t in self.threads]

    """Returns True if all threads are waiting for work"""

    def idle(self):
        return False not in [_i.is_set() for _i in self.idles]

    """Returns True if not tasks are left to be completed"""

    def done(self):
        return self.queue.empty()

    """Get the set of results that have been processed, repeatedly call until done"""

    def results(self, wait=0):
        sleep(wait)
        results = []
        # noinspection PyBroadException
        try:
            while True:
                # get a result, raises empty exception immediately if none available
                results.append(self.resultQueue.get(False))
                self.resultQueue.task_done()
        except:
            pass
        return results

    # run the test


if __name__ == '__main__':
    def work(x):
        sleep(5)
        logger.debug('Threadpool finished - main', id=x)
        return x


    def wait_for_results(_pool):
        while not _pool.done() or not _pool.idle():
            for _result in _pool.results():
                logger.debug("Threadpool got results - main", result=str(_result))
        for _result in _pool.results():
            logger.debug("Threadpool got result - main", result=str(_result))

    p = DiscoveryPool(3)
    logger.debug("Threadpool queueing work - main")
    for i in range(10):
        p.enqueue(work, i)
    logger.debug("Threadpool starting work - main")
    p.run(True)
    logger.debug("Threadpool waiting - main")
    sleep(8)
    for result in p.results():
        logger.debug("Threadpool get result - main", result=str(result))

    logger.debug("Threadpool cancel remaining work - main")
    p.abort()
    logger.debug("Threadpool waiting and restarting - main")
    p.run(True)
    logger.debug("Threadpool restarted waiting for results - main")
    wait_for_results(p)
    logger.debug("Threadpool adding more work - main")
    for i in range(10, 17):
        p.enqueue(work, i)
    wait_for_results(p)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import sys
import subprocess

import pkg_resources
from setuptools import setup
from setuptools import find_packages

# pypi publishing
# 1. set $HOME/.pypirc
#      [distutils]
#      index-servers =
#          pypi
#
#      [pypi]
#      username: <name>
#      password: <password>
# 2. deactivate  // if there's an active env
# 3. cd pycharmenv3; source bin/activate
# 4. pip3 install --upgrade wheel setuptools twine
# 5. cd <whatever_to>/agent
# 6. rm -rf dist/*
# 7. python3 setup.py sdist bdist_wheel
# 7a.twine check dist/*
# 8. twine upload dist/<latest>.tar.gz
# 9. you can test it with "pip3 install --upgrade --no-cache-dir foxyproxy"

VERSION = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

# Please update tox.ini when modifying dependency version requirements
install_requires = [
    'requests',
    'setuptools>=1.0',
    'logbook',
    'dnspython',
    'ldap3',
    'psutil'
]

if sys.platform != 'darwin':
    pass
if sys.version_info < (3, 6):
    sys.stderr.write("Python 2 and Python 3.5 or lower are not supported")
    exit(-1)

# crate a git_version file
# noinspection PyBroadException
try:
    git_version_b = subprocess.check_output(['git', 'describe', '--always'], stderr=subprocess.DEVNULL)  # returns bytes
    VERSION = git_version_b.decode().strip()

    with open('./keychest_agent/version.py', 'w') as git_file:
        git_file.write("\nclass Version(object):\n    VERSION = '" + VERSION + "'\n")
except:
    sys.stderr.write("You need to have installed 'git' to run the installer so it correctly determines its version")
    VERSION = "undetermined"

dev_extras = [
    'nose',
    'pep8',
    'tox'
]

docs_extras = [
    'Sphinx>=1.0',  # autodoc_member_order = 'bysource', autodoc_default_flags
    'sphinx_rtd_theme',
    'sphinxcontrib-programoutput'
]


setup(
    name='keychest_agent',
    version=VERSION,
    description='Keychest agent',
    url='https://gitlab.com/keychest/agent',
    author="Smart Arcs Ltd",
    author_email='support@keychest.net',
    long_description=open('README.md', encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security'
    ],

    packages=find_packages(),
    namespace_packages=['keychest_agent'],
    package_data={'keychest_agent': ['LICENSE']},
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'dev': dev_extras,
        'docs': docs_extras,
    },

    entry_points={
        'console_scripts': [
            'keychest_agent = keychest_agent.main:main'
        ],
    }
)
